#include "mus.h"

void StrToLetters(const string &s, vector<byte> &Letters);

extern ScoreMx g_SubstMx;
bool g_LogTB = false;
bool g_LogDP = false;
RefPairData *g_ptrRef;
static FILE *g_fOut;

void GetSubVec(const vector<byte> &x, uns From, uns To, vector<byte> &y)
	{
	y.clear();
	y.reserve(To - From + 1);
	uns L = SIZE(x);
	asserta(From <= To);
	asserta(From < L);
	asserta(To < L);
	for (uns i = From; i <= To; ++i)
		y.push_back(x[i]);
	}

//static SCORE AnchorAlnFunc(const vector<byte> &A, const vector<byte> &B,
//  string &Path)
//	{
//	const ScoreMx &GetB70C();
//	// const ScoreMx Mx = GetB70C();
//
//	SCORE NWLinear(const vector<byte> &A_, const vector<byte> &B_, const ParamVec &/*Params*/,
//	  string &Path);
//
//	ParamVec DummyParams;
//	return NWLinear(A, B, DummyParams, Path);
//	}

static double AlnPairMemSave(const ParamVec &Params, const PWSet &Set,
  uns PairIndex)
	{
	asserta(PairIndex < SIZE(Set.m_Names));
	const string &Name = Set.m_Names[PairIndex];

	uns IndexA = 2*PairIndex;
	uns IndexB = IndexA + 1;
	asserta(IndexB < SIZE(Set.m_GappedSeqs));

	const string &RefGSA = Set.m_GappedSeqs[IndexA];
	const string &RefGSB = Set.m_GappedSeqs[IndexB];

	const string &LabelA = Set.m_Labels[IndexA];
	const string &LabelB = Set.m_Labels[IndexB];

	vector<string> Labels;
	Labels.push_back(LabelA);
	Labels.push_back(LabelB);

	vector<byte> UVA;
	vector<byte> UVB;

	string USA;
	string USB;
	StripGaps(RefGSA, USA);
	StripGaps(RefGSB, USB);

	StrToLetters(USA, UVA);
	StrToLetters(USB, UVB);

	//const uns LA = SIZE(UVA);
	//const uns LB = SIZE(UVB);

	string Path;
	SCORE Score = g_PWA->NW(UVA, UVB, Params, false,
	  false, false, false, false, Path);

	UVA.clear();
	UVB.clear();

	vector<string> TestPWA(2);
	AlnStr(USA, Path, true, TestPWA[0]);
	AlnStr(USB, Path, false, TestPWA[1]);

	vector<string> RefPWA;
	RefPWA.push_back(RefGSA);
	RefPWA.push_back(RefGSB);

	const string &TestA = TestPWA[0];
	const string &TestB = TestPWA[1];

	double Q;
	double TC;
	QScore(Name.c_str(), TestPWA, Labels, RefPWA, Labels, Q, TC, false);

	if (g_LogAlns)
		{
		Log("\n");
		Log("%s\n", TestA.c_str());
		Log("%s\n", TestB.c_str());
		Log("%s\n", Path.c_str());
		Log("%s Q=%5.3f Score=%g\n", Name.c_str(), Q, Score);
		}
	return Q;
	}

static double AlnPair(const ParamVec &Params, const PWSet &Set, uns PairIndex,
  bool DoLoss)
	{
	if (Set.m_MemSave)
		return AlnPairMemSave(Params, Set, PairIndex);

	RefPairData Ref;
	Set.GetRefPair(PairIndex, Ref);
	g_ptrRef = &Ref;

	string Path;
	SCORE Score = 0;
	Score = g_PWA->NW(*Ref.UVA, *Ref.UVB, Params, DoLoss, 
	  Ref.ALeftTrunc, Ref.ARightTrunc, Ref.BLeftTrunc, Ref.BRightTrunc, Path);

	vector<string> TestPWA(2);
	AlnStr(*Ref.USA, Path, true, TestPWA[0]);
	AlnStr(*Ref.USB, Path, false, TestPWA[1]);

	const string &TestA = TestPWA[0];
	const string &TestB = TestPWA[1];

	if (g_fOut)
		{
		//double Weight = Set.GetWeight(PairIndex);
		const string &Name = Set.GetPairName(PairIndex);
		fprintf(g_fOut, "<%s\n", Name.c_str());
		fprintf(g_fOut, ">%s\n", Set.m_Labels[2*PairIndex].c_str());
		WriteFasta(g_fOut, TestA.c_str(), SIZE(TestA));
		fprintf(g_fOut, ">%s\n", Set.m_Labels[2*PairIndex+1].c_str());
		WriteFasta(g_fOut, TestB.c_str(), SIZE(TestB));
		}

	double Q;
	double TC;
	QScore(Ref.GetName().c_str(), TestPWA, *Ref.Labels, *Ref.RefPWA, *Ref.Labels,
	  Q, TC, false);

	if (g_LogAlns)
		{
		const string &Name = Ref.GetName();
		Log("\n");
		Log("%s\n", TestA.c_str());
		Log("%s\n", TestB.c_str());
		Log("%s\n", Path.c_str());
		Log("%s Q=%5.3f Score=%g\n", Name.c_str(), Q, Score);
		}
	return Q;
	}

void AlnPWSet()
	{
	const char *InFileName = RequiredValueOpt("alnpwset");
	const char *OutFileName = ValueOpt("out");
	bool DoLoss = FlagOpt("loss");
	g_LogAlns = FlagOpt("logalns");
	g_LogDP = FlagOpt("logdp");
	g_LogTB = FlagOpt("logtb");
	//bool DoDump = FlagOpt("dump");
	bool MemSave = FlagOpt("memsave");

	if (OutFileName != 0)
		g_fOut = CreateStdioFile(OutFileName);

	ParamVec Params;
	PairWiseAligner::InitFromCmdLine(Params);
	GetSubstMxFromCmdLine();

	PWSet Set;
	MSASet M;
	M.FromFile(InFileName);
	M.GetAllPairs(Set, MemSave);
	M.Clear();
	
	Log("\n");
	Log("PWA=%s\n", g_PWA->GetName());
	Params.LogMe();
	Log("\n");

	const uns PairCount = Set.GetPairCount();
	double SumQ = 0;
	double SumWeights = 0;
	double SumQWeighted = 0;
	ProgressStart("Aligning");
	uns Click = 10;
	for (uns PairIndex = 0; PairIndex < PairCount; ++PairIndex)
		{
		double Weight = Set.GetWeight(PairIndex);
		SumWeights += Weight;
		double Q = AlnPair(Params, Set, PairIndex, DoLoss);
		SumQ += Q;
		SumQWeighted += Q*Weight;
		if (PairIndex%Click == 0 || PairIndex + 1 == PairCount)
			{
			SetProgressMsg("Q=%6.4f", SumWeights == 0 ? 0.0 : SumQWeighted/SumWeights);
			ProgressStep(PairIndex, PairCount);
			}
		}
	ProgressDone();

	Log("\n");
	Log("Q =%6.4f\n", SumQ/PairCount);
	Log("Qw=%6.4f\n", SumQWeighted/SumWeights);

	if (g_fOut != 0)
		CloseStdioFile(g_fOut);
	}
