#ifndef dpmem_h
#define dpmem_h

#include <vector>
#include "utils.h"
#include "mustypes.h"
#include "progress.h"

using namespace std;

typedef SCORE **ScoreMx;
typedef char **CharMx;

class DPMemData
	{
public:
	DPMemData()
		{
		LA = 0;
		LB = 0;
		}

	~DPMemData()
		{
		Clear();
		}

	uns GetNS() const
		{
		return (uns) S.size();
		}

	uns GetNC() const
		{
		return (uns) C.size();
		}

	SCORE **GetS(uns i)
		{
		if (i >= S.size())
			Die("DPMemData::GetS(%u), out of range", i);
		return S[i];
		}

	SCORE **GetS(uns i) const
		{
		if (i >= S.size())
			Die("DPMemData::GetS(%u), out of range");
		return S[i];
		}

	char **GetC(uns i)
		{
		if (i >= S.size())
			Die("DPMemData::GetC(%u), out of range");
		return C[i];
		}

	char **GetC(uns i) const
		{
		if (i >= S.size())
			Die("DPMemData::GetC(%u), out of range");
		return C[i];
		}

	void Clear()
		{
		const uns NS = (uns) S.size();
		const uns NC = (uns) C.size();

		for (uns i = 0; i < NS; ++i)
			for (uns j = 0; j <= LA; ++j)
				delete[] S[i][j];

		for (uns i = 0; i < NC; ++i)
			for (uns j = 0; j <= LA; ++j)
				delete[] C[i][j];

		S.clear();
		C.clear();

		LA = 0;
		LB = 0;
		}

	void Copy(const DPMemData &rhs)
		{
		Alloc(rhs.LA, rhs.LB, rhs.GetNS(), rhs.GetNC());

		const uns NS = (uns) S.size();
		const uns NC = (uns) C.size();
		for (uns i = 0; i < NS; ++i)
			{
			SCORE **Si = GetS(i);
			const ScoreMx rhsSi = rhs.GetS(i);
			for (uns j = 0; j <= LA; ++j)
				{
				for (uns k = 0; k <= LB; ++k)
					Si[j][k] = rhsSi[j][k];
				}
			}

		for (uns i = 0; i < NC; ++i)
			{
			char **Ci = GetC(i);
			const CharMx rhsCi = rhs.GetC(i);
			for (uns j = 0; j <= LA; ++j)
				{
				for (uns k = 0; k <= LB; ++k)
					Ci[j][k] = rhsCi[j][k];
				}
			}
		}

	void Alloc(uns NewLA, uns NewLB, uns NS, uns NC);

	bool Eq(const DPMemData &rhs, bool DoLoss = false,
	  const string &Str = "", bool DieOnFirstDiff = false) const;

	void SetSMx(uns i, SCORE s)
		{
		for (uns j = 0; j <= LA; ++j)
			{
			SCORE *Row = S[i][j];
			for (uns k = 0; k <= LB; ++k)
				Row[k] = s;
			}
		}

	void SetCMx(uns i, char c)
		{
		for (uns j = 0; j <= LA; ++j)
			{
			SCORE *Row = S[i][j];
			for (uns k = 0; k <= LB; ++k)
				Row[k] = c;
			}
		}

	void Set(SCORE s, char c)
		{
		uns NS = (uns) S.size();
		uns NC = (uns) C.size();

		for (uns i = 0; i < NS; ++i)
			SetSMx(i, s);

		for (uns i = 0; i < NC; ++i)
			SetCMx(i, c);
		}

public:
	uns LA;
	uns LB;
	uns LastLA;
	uns LastLB;

private:
	vector<SCORE **> S;
	vector<char **> C;
	};

#define NW_ALLOCDP(LA, LB)						\
	g_DPMem.Alloc(LA, LB, 4, 3);

#define GL_ALLOCDP(LA, LB)						\
	g_DPMem.Alloc(LA, LB, 8, 7);

#define NW_CKALLOCDP(LA, LB) g_DPMem.Alloc(LA, LB, 4, 3);

#define USW_CKALLOCDP(LA, LB) g_DPMem.Alloc(LA, LB, 4, 3);

#define DA_CKALLOCDP(LA, LB) g_DPMem.Alloc(LA, LB, 5, 5);

#define GL_CKALLOCDP(LA, LB)					\
	if (g_DPMem.LA < LA || g_DPMem.LB < LB ||	\
	  g_DPMem.GetNS() < 8 || g_DPMem.GetNC() < 7)\
		Die("GL_ALLOCMX");

#define USW_DEFMX()								\
	ScoreMx M = g_DPMem.GetS(0);				\
	CharMx TM = g_DPMem.GetC(0);

#define NW_DEFMX()								\
	ScoreMx Loss = g_DPMem.GetS(0);				\
	ScoreMx M = g_DPMem.GetS(1);				\
	ScoreMx D = g_DPMem.GetS(2);				\
	ScoreMx I = g_DPMem.GetS(3);				\
	CharMx TM = g_DPMem.GetC(0);				\
	CharMx TD = g_DPMem.GetC(1);				\
	CharMx TI = g_DPMem.GetC(2);

#define DA_DEFMX()								\
	ScoreMx M = g_DPMem.GetS(0);				\
	ScoreMx D = g_DPMem.GetS(1);				\
	ScoreMx I = g_DPMem.GetS(2);				\
	ScoreMx X = g_DPMem.GetS(3);				\
	ScoreMx Y = g_DPMem.GetS(4);				\
	CharMx TM = g_DPMem.GetC(0);				\
	CharMx TD = g_DPMem.GetC(1);				\
	CharMx TI = g_DPMem.GetC(2);				\
	CharMx TX = g_DPMem.GetC(3);				\
	CharMx TY = g_DPMem.GetC(4);

#define GL_DEFMX()								\
	ScoreMx Loss = g_DPMem.GetS(0);			\
	ScoreMx MA = g_DPMem.GetS(1);				\
	ScoreMx MB = g_DPMem.GetS(2);				\
	ScoreMx MC = g_DPMem.GetS(3);				\
	ScoreMx DA = g_DPMem.GetS(4);				\
	ScoreMx DC = g_DPMem.GetS(5);				\
	ScoreMx IA = g_DPMem.GetS(6);				\
	ScoreMx IC = g_DPMem.GetS(7);				\
	CharMx TMA = g_DPMem.GetC(0);				\
	CharMx TMB = g_DPMem.GetC(1);				\
	CharMx TMC = g_DPMem.GetC(2);				\
	CharMx TDA = g_DPMem.GetC(3);				\
	CharMx TDC = g_DPMem.GetC(4);				\
	CharMx TIA = g_DPMem.GetC(5);				\
	CharMx TIC = g_DPMem.GetC(6);

#endif // dpmem_h
