#include "mus.h"
#include "hillclimber.h"

bool g_LogAlns;
static PWSet *g_ptrPWSet;
//static MSASet *g_ptrMSASet;

static double PWGetQ(const ParamVec &Params)
	{
	return g_ptrPWSet->ComputeQ(Params);
	}

void HillClimb()
	{
	const char *InFileName = RequiredValueOpt("hillclimb");
	//bool DoMSAs = FlagOpt("msa");
	g_LogAlns = FlagOpt("logalns");
	unsigned SubsetSize = UINT_MAX;
	IntOpt("subset", &SubsetSize);
	bool Fast = FlagOpt("fast");

	unsigned MaxDirs = UINT_MAX;
	double Mindq = 0.001;
    double Maxdq = 0.01;
    double Mindp = 0.01;
    double Maxdp = 0.2;

	IntOpt("maxdirs", &MaxDirs);
	FloatOpt("mindq", &Mindq);
	FloatOpt("maxdq", &Maxdq);
	FloatOpt("mindp", &Mindp);
	FloatOpt("maxdp", &Maxdp);

	ParamVec Params;
	PairWiseAligner::InitFromCmdLine(Params);
	GetSubstMxFromCmdLine();
	ParamVec::FixMF(true);

	MSASet M;
	PWSet Set;
	M.FromFile(InFileName);
	M.GetAllPairs(Set);
	M.Clear();

	g_ptrPWSet = &Set;

	Log("\n");
	Log("Initial params: ");
	Params.LogMe();
	Log("\n");

	ParamVec Gradients;
	Gradients.Zero(0.0);

	HillClimber H;
	H.m_MindQ = Mindq;
	H.m_MaxdQ = Maxdq;
	H.m_MindParam = Mindp;
	H.m_MaxdParam = Maxdp;
	H.m_Fast = Fast;

	H.Climb(PWGetQ, Params, Gradients, MaxDirs);

	string tmp1;
	string tmp2;
	Log("Algo=hc;Set=%s;Mx=%s;Q=%7.5f;%sCmdLine=%s\n",
	  InFileName,
	  RequiredValueOpt("mx"),
	  H.m_BestQ,
	  H.m_BestParams.ToSepStr(tmp1).c_str(),
	  GetCmdLine(tmp2).c_str());
	}
