#include "mus.h"

HillClimber::HillClimber()
	{
	Clear();
	}

void HillClimber::Clear()
	{
	m_GetQFn = 0;
	m_BestParams.Zero(1.0);
	m_BestQ = 0;
	m_Qs.clear();
	m_Params.clear();
	m_DirsVec.clear();
	m_MindQ = 0.001;
	m_MaxdQ = 0.01;
	m_MinGradient = 0.0001;
	m_MindParam = 0.0001;
	m_MaxdParam = 1.0;
	m_Fast = false;
	m_StepCount = 0;
	m_ImprovedBothPlusAndMinusCount = 0;
	m_MultipleAxisCount = 0;
	m_NotImprovedDirCount = 0;
	}

// Enumerate combinations of k objects from N
void InitCombination(vector<uns> &Comb, uns k)
	{
	asserta(k > 0);
	Comb.clear();
	for (uns i = 0; i < k; ++i)
		Comb.push_back(i);
	}

bool GetNextCombination(vector<uns> &Comb, uns N)
	{
	uns k = SIZE(Comb);
	asserta(k > 0);
	asserta(N >= k);

	int i = k - 1;
    while (i >= 0 && Comb[i] >= N - k + i)
		--i;

	if (i < 0)
		return false;

	++(Comb[i]);
    for (uns m = i + 1; m < k; ++m)
        Comb[m] = Comb[m-1] + 1;
    return true;
	}

// Return all possible direction vectors of length N
// where N is the number of alignment parameters and
// each value is +1, 0 or -1.
static void GetAllDirs(vector<vector<int> > &DirsVec, uns N)
	{
	DirsVec.clear();
	vector<int> Dir;
	for (uns ParamIndex = 0; ParamIndex < N; ++ParamIndex)
		Dir.push_back(-1);

	for (;;)
		{
		DirsVec.push_back(Dir);
		bool Updated = false;
		for (uns ParamIndex = 0; ParamIndex < N; ++ParamIndex)
			{
			if (Dir[ParamIndex] != 1)
				{
				++Dir[ParamIndex];
				for (uns ParamIndex2 = 0; ParamIndex2 < ParamIndex;
				  ++ParamIndex2)
					Dir[ParamIndex2] = -1;
				Updated = true;
				break;
				}
			}
		if (!Updated)
			break;
		}
	}

void LogDirsVec(const vector<vector<int> > &DirsVec)
	{
	const uns N = SIZE(DirsVec);
	for (uns i = 0; i < N; ++i)
		{
		const vector<int> &v = DirsVec[i];
		const uns M = SIZE(v);
		for (uns j = 0; j < M; ++j)
			Log(j == 0 ? "%3d" : " %3d", v[j]);
		Log("\n");
		}
	}

static void GetDirs(vector<vector<int> > &DirsVec, uns MaxNonZeroDirs)
	{
	DirsVec.clear();
	set<vector<int> > DirsSet;

	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	if (MaxNonZeroDirs > ActiveParamCount)
		MaxNonZeroDirs = ActiveParamCount;

	vector<uns> NonZeroDirParamIndexes;
	InitCombination(NonZeroDirParamIndexes, MaxNonZeroDirs);

	do
		{
		vector<vector<int> > NonZeroDirsVec;
		GetAllDirs(NonZeroDirsVec, MaxNonZeroDirs);

		vector<int> Dirs(ActiveParamCount, 0);

		const uns N = SIZE(NonZeroDirsVec);
		for (uns i = 0; i < N; ++i)
			{
			const vector<int> &NonZeroDirs = NonZeroDirsVec[i];
			const uns M = SIZE(NonZeroDirParamIndexes);
			for (uns j = 0; j < M; ++j)
				{
				uns ParamIndex = NonZeroDirParamIndexes[j];
				Dirs[ParamIndex] = NonZeroDirs[j];
				}
			DirsSet.insert(Dirs);
			}
		}
	while (GetNextCombination(NonZeroDirParamIndexes, ActiveParamCount));

	for (set<vector<int> >::const_iterator p = DirsSet.begin();
	  p != DirsSet.end(); ++p)
		DirsVec.push_back(*p);
	}

static void DirsToStr(const vector<int> &Dirs, const ParamVec &PlusDeltas,
  const ParamVec &MinusDeltas, string &Why)
	{
	Why.clear();
	bool AllZero = true;
	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	asserta(SIZE(Dirs) == ActiveParamCount);
	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		char s[32];
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		switch (Dirs[ActiveParamIndex])
			{
		case -1:
			sprintf(s, "-%s(%g)", AlnParamToStr(ParamIndex),
			  MinusDeltas.Get(ParamIndex));
			if (!Why.empty())
				Why += " ";
			Why += s;
			AllZero = false;
			break;
		case 0:
			break;
		case 1:
			sprintf(s, "+%s(%g)", AlnParamToStr(ParamIndex),
			  PlusDeltas.Get(ParamIndex));
			if (!Why.empty())
				Why += " ";
			Why += s;
			AllZero = false;
			break;
			}
		}
	if (AllZero)
		Why = "(center)";
	}

void HillClimber::LogGetQ(const ParamVec &Params, double Q,
  const string &Why, bool CacheHit) const
	{
	m_BestParams.LogRow();
	Log("  %7.5f", m_BestQ);
	m_Gradients.LogRow();

	Params.LogRow();
	Log("  %7.5f", Q);
	Log("%c%c", Q == m_BestQ ? '*' : ' ', CacheHit ? '!' : ' ');
	Log("  %s", Why.c_str());
	Log("\n");

	Progress("BestQ=%7.5f Q=%7.5f%s%s %s",
	  m_BestQ, Q, Q == m_BestQ ? "*" : "", CacheHit ? "!" : "", Why.c_str());
	}

void HillClimber::LogGetQHdr() const
	{
	ParamVec::LogRowHdr1('*');
	Log("       Q*");
	ParamVec::LogRowHdr1('\'');
	ParamVec::LogRowHdr1();
	Log("        Q");
	Log("\n");

	ParamVec::LogRowHdr2();
	Log("  -------");
	ParamVec::LogRowHdr2();
	ParamVec::LogRowHdr2();
	Log("  -------");
	Log("\n");
	}

// InitParams[ParamIndex] += Delta gives an improvement, search close by
// for even better improvement.
SCORE HillClimber::GetImprovedDeltaParam(double StartQ, const ParamVec &InitParams,
  uns ParamIndex, SCORE InitDelta, bool Plus, double &NewQ)
	{
	SCORE BestDelta = InitDelta;

// Note "best" here means only best exploring in one direction (plus or minus)
// on one axis; a better improvement may have already been achieved on another
// axis.
	NewQ = StartQ;
	for (;;)
		{
		bool ImprovedThisRound = false;
	// Try with smaller values k Delta/4, Delta/2, 3 Delta/4
		for (int i = 1; i <= 3; ++i)
			{
			SCORE Delta = (SCORE) (i*BestDelta/4.0);
			ParamVec Params(InitParams);
			SCORE Param = Params.Get(ParamIndex);
			if (Plus)
				Param += Delta;
			else
				Param -= Delta;
			Params.Set(ParamIndex, Param);

			char Why[64];
			sprintf(Why, "%d<< %s(%c%g)", i, AlnParamToStr(ParamIndex), pom(Plus), Delta);
			double Q = GetQ(Params, Why);
			if (Q > NewQ)
				{
				double dQ = fabs(Q - StartQ);
				BestDelta = Delta;
				ImprovedThisRound = true;
				NewQ = Q;
				if (dQ > m_MindQ)
					{
					Log("%cd%s dQ=%g > min=%g, return (from smaller loop).\n",
					  pom(Plus), AlnParamToStr(ParamIndex), dQ, m_MindQ);
					return BestDelta;
					}
				}
			}
		if (!ImprovedThisRound)
			break;
		}

// Try with larger values 5 Delta/4 3 Delta/2 2 Delta
	for (int i = 5; i <= 8; ++i)
		{
		SCORE Delta = (SCORE) (i*BestDelta/4.0);
		ParamVec Params(InitParams);
		SCORE Param = Params.Get(ParamIndex);
		if (Plus)
			Param += Delta;
		else
			Param -= Delta;
		Params.Set(ParamIndex, Param);

		char Why[64];
		sprintf(Why, "%d>> %s(%c%g)", i, AlnParamToStr(ParamIndex), pom(Plus), Delta);
		double Q = GetQ(Params, Why);
		if (Q > NewQ)
			{
			double dQ = fabs(Q - StartQ);
			NewQ = Q;
			BestDelta = Delta;
			if (dQ > m_MindQ)
				{
				Log("%cd%s dQ=%g > min=%g, return (from bigger loop).\n",
				  pom(Plus), AlnParamToStr(ParamIndex), dQ, m_MindQ);
				return BestDelta;
				}
			}
		}
	return BestDelta;
	}

SCORE HillClimber::GetDeltaParam(double CenterQ, const ParamVec &InitParams,
  uns ParamIndex, bool Plus, bool *ptrImproved)
	{
	if (ptrImproved != 0)
		*ptrImproved = false;
	const char *Name = AlnParamToStr(ParamIndex);
	SCORE Gradient = m_Gradients.Get(ParamIndex);
	double TargetdQ = (m_MindQ + m_MaxdQ)/2;
	SCORE dParam = (SCORE) (TargetdQ/Gradient);
	if (ParamIndex == P_b || ParamIndex == P_h)
		dParam = SCORE(0.2);//@@
	else if (ParamIndex == P_de)
		dParam = SCORE(0.02);
	SCORE BestdParam = SCORE((m_MindParam + m_MaxdParam)/2);
	set<double> dParams;
	char Sign = pom(Plus);
	Log("GetDeltaParam(%c%s)\n", Sign, Name);

	bool Maxed = false;
	bool Mined = false;
	SCORE LastMindParam = (SCORE) m_MindParam;
	SCORE LastMaxdParam = (SCORE) m_MaxdParam;
	for (uns Count = 0; Count < 8; ++Count)
		{
		if (dParam < m_MindParam)
			dParam = (SCORE) m_MindParam;
		if (dParam > m_MaxdParam)
			dParam = (SCORE) m_MaxdParam;

		ParamVec Params(InitParams);
		SCORE Param = Params.Get(ParamIndex) + (Plus ? dParam : -dParam);
		Params.Set(ParamIndex, Param);

		if (dParams.find(dParam) != dParams.end())
			{
			Log("%cd(%s)=%g already tried, give up\n", Sign, Name, dParam);
			break;
			}
		dParams.insert(dParam);

		char Why[64];
		sprintf(Why, "%cd%s(%g=%g)", Sign, Name, dParam, Param);
		double Q = GetQ(Params, Why);
		if (Q > CenterQ)
			{
			if (ptrImproved != 0)
				*ptrImproved = true;
			if (m_Fast)
				return dParam;
			double NewQ = CenterQ;
			Log("GetImprovedDeltaParam(%c%s)\n", Sign, Name);
			dParam = GetImprovedDeltaParam(CenterQ, InitParams, ParamIndex,
			  dParam, Plus, NewQ);
			Gradient = SCORE(fabs(NewQ - Q)/dParam);
			m_Gradients.Set(ParamIndex, Gradient);
			return dParam;
			}

		double dQ = fabs(CenterQ - Q);
		if (dQ > 0 && dQ < 2*m_MaxdQ)
			BestdParam = dParam;
		if (dQ == 0)
			{
			Log("%cd(%s)=%g, dQ=0", Sign, Name, dParam);
			dParam *= 4;
			Log(", increase to %g\n", dParam);
			continue;
			}
		else if (dQ < m_MindQ)
			{
			if (Maxed)
				{
				Log("%cd(%s)=%g dQ=%g", Sign, Name, dParam, dQ);
			//	dParam = (LastMaxdParam + LastMindParam)/2;
				Log(", oscillating max-min, set to min=%g\n", LastMindParam);
				return dParam;
				}
			Mined = true;
			LastMindParam = dParam;
			Log("%cd(%s)=%g", Sign, Name, dParam);
			dParam *= 2;
			Log(" dQ=%g < min=%g, increase to %g\n", dQ, m_MindQ, dParam);
			}
		else if (dQ > m_MaxdQ)
			{
			if (Mined)
				{
				Log("%cd(%s)=%g dQ=%g", Sign, Name, dQ);
				dParam = LastMindParam;
				Log(", oscillating min-max, set to min=%g\n", LastMindParam);
				return dParam;
				}
			Maxed = true;
			LastMaxdParam = dParam;
			Log("%cd(%s)=%g", Sign, Name, dParam);
			dParam = (SCORE) (dParam*TargetdQ/dQ);
			Log(" dQ=%g > max=%g, decrease to %g\n", dQ, m_MaxdQ, dParam);
			}
		else
			{
			Log("%cd(%s)=%g dQ=%g, accepted\n", Sign, Name, dParam, dQ);
			return dParam;
			}
		}

	if (BestdParam > 0.2)
		{
		if (ParamIndex == P_b || ParamIndex == P_h ||
		  ParamIndex == P_e || ParamIndex == P_te)
			return SCORE(0.2);//@@
		}
	Log("Gave up finding %cd(%s), return best=%g\n", Sign, Name, BestdParam);
	return BestdParam;
	}

void HillClimber::GetParamDeltas(double CenterQ, const ParamVec &Params,
  ParamVec &PlusDeltas, ParamVec &MinusDeltas)
	{
	PlusDeltas.Zero(0);
	MinusDeltas.Zero(0);

	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);

		bool ImprovedPlus = false;
		bool ImprovedMinus = false;

		SCORE DeltaParamPlus =
		  GetDeltaParam(CenterQ, Params, ParamIndex, true, &ImprovedPlus);

		SCORE DeltaParamMinus =
		  GetDeltaParam(CenterQ, Params, ParamIndex, false, &ImprovedMinus);

		PlusDeltas.Set(ParamIndex, DeltaParamPlus);
		MinusDeltas.Set(ParamIndex, DeltaParamMinus);

		m_ImprovedPlus[ActiveParamIndex] = ImprovedPlus;
		m_ImprovedMinus[ActiveParamIndex] = ImprovedMinus;
		}
	}

static void DeltaParams(const ParamVec &StartParams, const vector<int> &Dirs,
  const ParamVec &PlusDeltas, const ParamVec &MinusDeltas, ParamVec &Params)
	{
	Params = StartParams;
	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		SCORE Param = Params.Get(ParamIndex);
		asserta(ActiveParamIndex < SIZE(Dirs));
		switch (Dirs[ActiveParamIndex])
			{
		case -1:
			{
			SCORE Delta = MinusDeltas.Get(ParamIndex);
			Params.Set(ParamIndex, Param - Delta);
			break;
			}
		case 0:
			break;
		case 1:
			{
			SCORE Delta = PlusDeltas.Get(ParamIndex);
			Params.Set(ParamIndex, Param + Delta);
			break;
			}
		default:
			asserta(false);
			}
		}
	}

bool HillClimber::FastIter(const ParamVec &StartParams)
	{
	double StartQ = GetQ(StartParams, "Iter center");

	bool Improved = false;
	m_BestParams = StartParams;
	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);

		bool PlusFirst = (rand() > RAND_MAX/2);
		double QBefore = m_BestQ;
		GetDeltaParam(StartQ, m_BestParams, ParamIndex, PlusFirst);
		if (m_BestQ > QBefore)
			{
			Improved = true;
			continue;
			}
		GetDeltaParam(StartQ, m_BestParams, ParamIndex, !PlusFirst);
		if (m_BestQ > QBefore)
			Improved = true;
		}
	return Improved;
	}

bool HillClimber::Iter(const ParamVec &StartParams)
	{
	if (m_Fast)
		return FastIter(StartParams);

	const uns ActiveParamCount = ParamVec::GetActiveParamCount();

	double StartQ = GetQ(StartParams, "Iter center");

	m_ImprovedPlus.clear();
	m_ImprovedMinus.clear();
	m_ImprovedPlus.resize(ActiveParamCount, false);
	m_ImprovedMinus.resize(ActiveParamCount, false);

	ParamVec PlusDeltas;
	ParamVec MinusDeltas;

	Log("\n");
	Log("Find deltas:\n");
	GetParamDeltas(StartQ, StartParams, PlusDeltas, MinusDeltas);

	Log("\n");
	Log("Hypercuboid: ");

	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		SCORE Param = StartParams.Get(ParamIndex);
		SCORE MinusDelta = MinusDeltas.Get(ParamIndex);
		SCORE PlusDelta = PlusDeltas.Get(ParamIndex);
		Log(" %s:%.5g,+%.5g,-%.5g;", AlnParamToStr(ParamIndex), Param, PlusDelta, MinusDelta);
		}
	Log("\n");

	Log("\n");
	Log("Explore hypercuboid:\n");
	const uns DirCount = SIZE(m_DirsVec);
	uns BestDir = UINT_MAX;
	double BestDirQ = StartQ;
	bool Improved = false;
	for (uns DirIndex = 0; DirIndex < DirCount; ++DirIndex)
		{
		const vector<int> &Dirs = m_DirsVec[DirIndex];

		ParamVec Params;
		DeltaParams(StartParams, Dirs, PlusDeltas, MinusDeltas, Params);
		string Why;
		DirsToStr(Dirs, PlusDeltas, MinusDeltas, Why);
		char Tmp[32];
		sprintf(Tmp, "Dir %u of %u ", DirIndex+1, DirCount);
		Why = Tmp + Why;
		double Q = GetQ(Params, Why);
		if (Q > m_BestQ)
			{
			m_BestParams = Params;
			m_BestQ = Q;
			}
		if (Q > BestDirQ)
			{
			Improved = true;
			BestDir = DirIndex;
			BestDirQ = Q;
			}
		}
	if (Improved)
		{
		StepStats(BestDir);
		m_Qs.push_back(m_BestQ);
		m_Params.push_back(m_BestParams);
		}
	return Improved;
	}

double HillClimber::GetQ(const ParamVec &Params, const string &Why)
	{
	map<ParamVec, double>::const_iterator p = m_ParamsToQ.find(Params);
	if (p != m_ParamsToQ.end())
		{
		++m_ParamsToQHitCount;
		double Q = p->second;
		LogGetQ(Params, Q, Why, true);
		return Q;
		}

	double Q = m_GetQFn(Params);
	if (Q > m_BestQ)
		{
		m_BestQ = Q;
		m_BestParams = Params;
		}
	LogGetQ(Params, Q, Why);
	m_ParamsToQ[Params] = Q;
	return Q;
	}

void HillClimber::StepStats(uns DirIndex)
	{
	++m_StepCount;
	asserta(DirIndex < SIZE(m_DirsVec));
	const vector<int> &Dirs = m_DirsVec[DirIndex];
	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	uns AxisCount = 0;
	Log("\n");
	Log("IPM: ");
	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		const uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		bool Plus = m_ImprovedPlus[ActiveParamIndex];
		bool Minus = m_ImprovedMinus[ActiveParamIndex];
		if (ActiveParamIndex > 0)
			Log(",");
		Log(" %s +%c -%c", AlnParamToStr(ParamIndex), tof(Plus), tof(Minus));
		}
	Log("\n");
	Log("Step: ");
	for (uns ActiveParamIndex = 0; ActiveParamIndex < ActiveParamCount;
	  ++ActiveParamIndex)
		{
		int Dir = Dirs[ActiveParamIndex];
		if (Dir == 0)
			continue;

		const uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		if (Dir == 1)
			Log(" +");
		else if (Dir == -1)
			Log(" -");
		else
			Log("??Dir=%d??", Dir);

		Log("%s", AlnParamToStr(ParamIndex));

		if ((Dir == 1 && !m_ImprovedPlus[ActiveParamIndex])
		  || (Dir == -1 && !m_ImprovedMinus[ActiveParamIndex]))
			Log("[NotImproved]");

		++AxisCount;
		if (Dir == 1 && !m_ImprovedPlus[ActiveParamIndex])
			++m_NotImprovedDirCount;
		else if (Dir == -1 && !m_ImprovedMinus[ActiveParamIndex])
			++m_NotImprovedDirCount;
		}
	Log("\n");
	asserta(AxisCount > 0);
	if (AxisCount > 0)
		++m_MultipleAxisCount;
	}

void HillClimber::Climb(ptrGetQFn GetQFn, const ParamVec &Params,
  const ParamVec &Gradients, uns MaxDirs, const string &Comment)
	{
	m_Qs.clear();
	m_Params.clear();
	m_ParamsToQ.clear();
	m_ParamsToQHitCount = 0;

	m_BestParams = Params;
	m_Gradients = Gradients;
	m_GetQFn = GetQFn;

	Log("\n");
	Log("Climb parameters:\n");
	Log("  MindQ       %10.5f\n", m_MindQ);
	Log("  MaxdQ       %10.5f\n", m_MaxdQ);
	Log("  MindParam   %10.5f\n", m_MindParam);
	Log("  MaxdParam   %10.5f\n", m_MaxdParam);
	Log("  MinGradient %10.5f\n", m_MinGradient);
	Log("  Fast        %10c\n", tof(m_Fast));
	if (MaxDirs == UINT_MAX)
		Log("  Max dirs           All\n");
	else
		Log("  Max dirs    %10u\n", MaxDirs);
	Log("\n");

	uns ActiveParamCount = ParamVec::GetActiveParamCount();
	if (MaxDirs == UINT_MAX)
		GetAllDirs(m_DirsVec, ActiveParamCount);
	else
		{
		if (MaxDirs > ActiveParamCount)
			{
			Warning("-maxdirs %u but only %u active params",
			  MaxDirs, ActiveParamCount);
			MaxDirs = ActiveParamCount;
			}
		GetDirs(m_DirsVec, MaxDirs);
		}

	SetLogProgress(false);
	LogGetQHdr();
	m_BestQ = GetQ(Params, "Hill climb start");

	m_Qs.push_back(m_BestQ);
	m_Params.push_back(Params);

	for (;;)
		{
		ParamVec StartParams(m_BestParams);
		bool Improved = Iter(StartParams);
		Log("END ITER: %s, BestQ=%7.5f\n",
		  Improved ? "improved" : "not improved", m_BestQ);
		Progress("END ITER: %s, Q=%7.5f",
		  Improved ? "improved" : "not improved", m_BestQ);
		if (!Improved)
			{
			if (m_Fast)
				{
				m_Fast = false;
				continue;
				}
			break;
			}
		}
	Log("\n");
	Log("Path:\n");
	const uns N = SIZE(m_Qs);
	for (uns i = 0; i < N; ++i)
		GetQ(m_Params[i]);

	Log("\n");
	Log("Climb done %s\n", Comment.c_str());
	Log("%10u  Steps\n", m_StepCount);
	Log("%10u  Multiple axis\n", m_MultipleAxisCount);
	Log("%10u  Improved both plus and minus\n", m_ImprovedBothPlusAndMinusCount);
	Log("%10u  Not improved dir\n", m_NotImprovedDirCount);

	Log("\n");
	Log("BestQ=%7.5f\n", m_BestQ);
	Log("BestParams=");
	m_BestParams.LogMe(true);
	Log("%u hits to Q cache\n", m_ParamsToQHitCount);
	}
