#ifndef hillclimber_h
#define hillclimber_h

#include "paramspace.h"

typedef double (*ptrGetQFn)(const ParamVec &Params);

class HillClimber
	{
public:
	HillClimber();
	void Clear();
	void LogGetQ(const ParamVec &Params, double Q,
	  const string &Why = "", bool CacheHit = false) const;
	void LogGetQHdr() const;
	double GetQ(const ParamVec &Params, const string &Why = "");
	void Climb(ptrGetQFn GetQFn, const ParamVec &InitParams,
	  const ParamVec &InitGradients, uns MaxDirs = UINT_MAX,
	  const string &Comment = "");
	bool Iter(const ParamVec &StartParams);
	bool FastIter(const ParamVec &StartParams);
	SCORE GetDeltaParam(double CenterQ, const ParamVec &InitParams,
	  uns ParamIndex, bool Plus, bool *ptrImproved = 0);
	SCORE GetImprovedDeltaParam(double CenterQ, const ParamVec &InitParams,
	  uns ParamIndex, SCORE InitDelta, bool Plus, double &NewQ);
	void GetParamDeltas(double CenterQ, const ParamVec &Params,
	  ParamVec &PlusDeltas, ParamVec &MinusDeltas);
	void StepStats(uns DirIndex);

public:
	ptrGetQFn m_GetQFn;
	ParamVec m_BestParams;
	ParamVec m_Gradients;
	double m_BestQ;
	vector<double> m_Qs;
	vector<ParamVec> m_Params;
	vector<vector<int> > m_DirsVec;
	map<ParamVec, double> m_ParamsToQ;
	vector<bool> m_ImprovedPlus;
	vector<bool> m_ImprovedMinus;
	unsigned m_ParamsToQHitCount;
	double m_MindQ;
	double m_MaxdQ;
	double m_MinGradient;
	double m_MindParam;
	double m_MaxdParam;
	bool m_Fast;

	uns m_StepCount;
	uns m_ImprovedBothPlusAndMinusCount;
	uns m_MultipleAxisCount;
	uns m_NotImprovedDirCount;
	};

#endif // hillclimber_h

