#include "mus.h"

static char g_LetterToChar[256];
static byte g_CharToLetter[256];
static unsigned g_AlphabetSize;
static byte g_GapLetter;
static bool g_InitDone = false;

bool InitLetters()
	{
	if (g_InitDone)
		return true;

	unsigned i = 0;

	for (unsigned i = 0; i < 256; ++i)
		{
		g_CharToLetter[i] = INVALID_LETTER;
		g_LetterToChar[i] = '?';
		}

#define X(c)										\
	{												\
	g_LetterToChar[i] = c;							\
	g_CharToLetter[toupper((unsigned char) c)] = i;	\
	g_CharToLetter[tolower((unsigned char) c)] = i;	\
	++i;											\
	}

// Amino acids
	X('A')
	X('C')
	X('D')
	X('E')
	X('F')
	X('G')
	X('H')
	X('I')
	X('K')
	X('L')
	X('M')
	X('N')
	X('P')
	X('Q')
	X('R')
	X('S')
	X('T')
	X('V')
	X('W')
	X('Y')

// Wild-cards
	X('X')
	X('B')
	X('Z')

// Stop codon
	X('*')

// Special case for gap symbol
	g_GapLetter = i;
	g_CharToLetter[(unsigned char) '-'] = g_GapLetter;
	g_CharToLetter[(unsigned char) '.'] = g_GapLetter;
	g_LetterToChar[g_GapLetter] = '-';

	g_AlphabetSize = g_GapLetter + 1;

	g_InitDone = true;
	return true;
	}
static bool InitDone = InitLetters();

unsigned GetAlphabetSize()
	{
	return g_AlphabetSize;
	}

byte CharToLetter(char c)
	{
	byte Letter = g_CharToLetter[(unsigned char) c];
	if (Letter == INVALID_LETTER)
		{
		if (isprint(c))
			Die("Invalid char %c", c);
		else
			Die("Invalid char 0x%02x", (byte) c);
		}
	return Letter;
	}

byte CharToLetterX(char c)
	{
	return g_CharToLetter[(unsigned char) c];
	}

char LetterToChar(byte Letter)
	{
	return g_LetterToChar[Letter];
	}

bool IsGapChar(char c)
	{
	return g_CharToLetter[(unsigned char) c] == g_GapLetter;
	}

bool IsGapLetter(byte b)
	{
	return b == g_GapLetter;
	}

void StrToLetters(const string &s, vector<byte> &Letters)
	{
	Letters.clear();
	Letters.reserve(s.size());
	for (string::const_iterator p = s.begin(); p != s.end(); ++p)
		{
		char c = *p;
		byte Letter = CharToLetter(c);
		Letters.push_back(Letter);
		}
	}

void StrToLettersX(const string &s, vector<byte> &Letters)
	{
	Letters.clear();
	Letters.reserve(s.size());
	for (string::const_iterator p = s.begin(); p != s.end(); ++p)
		{
		char c = *p;
		byte Letter = CharToLetterX(c);
		Letters.push_back(Letter);
		}
	}

void StripGaps(const vector<byte> &s, vector<byte> &t)
	{
	t.clear();
	t.reserve(s.size());
	for (vector<byte>::const_iterator p = s.begin(); p != s.end(); ++p)
		{
		byte b = *p;
		if (!IsGapLetter(b))
			t.push_back(b);
		}
	}

void StripGaps(const string &s, string &t)
	{
	t.clear();
	t.reserve(s.size());
	for (string::const_iterator p = s.begin(); p != s.end(); ++p)
		{
		char b = *p;
		if (!IsGapChar(b))
			t.push_back(b);
		}
	}

uns UngappedSeqLength(const string &s)
	{
	string t;
	StripGaps(s, t);
	return SIZE(t);
	}

void LettersToStr(const vector<byte> &Letters, string &Seq)
	{
	Seq.clear();
	const uns L = SIZE(Letters);
	for (uns i = 0; i < L; ++i)
		{
		byte Letter = Letters[i];
		char c = LetterToChar(Letter);
		Seq.push_back(c);
		}
	}
