#include "mus.h"

/***
Let test be an alignment to be evaluated and ref the true (reference)
alignment.

Let L be the number of columns in ref and d=1/L.

Define the loss function L(test, ref) = 1 - Q, where Q is the quality
score Q = (nr correct columns)/L.

L is a measure of the distance between test and ref, it is 0 if test=ref,
1 if there are no columns in common, some intermediate value otherwise.

Loss[i][j] is -d if A[i],B[j] are aligned in ref, 0 otherwise.
***/

float **g_Loss;
static unsigned g_MaxL = 0;

void SetLoss(const string &A, const string &B, double Scale)
	{
	uns GLA = (uns) A.size();
	uns GLB = (uns) B.size();

	uns MaxL = max(GLA, GLB);
	if (MaxL > g_MaxL)
		{
		if (g_MaxL > 0)
			{
			for (unsigned i = 0; i < g_MaxL; ++i)
				delete[] g_Loss[i];
			delete[] g_Loss;
			}

		g_MaxL = MaxL + 128;
		g_Loss = new float*[g_MaxL];
		for (unsigned i = 0; i < g_MaxL; ++i)
			g_Loss[i] = new float[g_MaxL];
		}

	const int L = (int) A.size();
	if ((int) B.size() != L)
		Die("SetLoss, A, B have different lengths");

	int LA = 0;
	int LB = 0;
	int N = 0;
	for (int p = 0; p < L; ++p)
		{
		char a = A[p];
		char b = B[p];
		if (!IsGapChar(a))
			++LA;
		if (!IsGapChar(b))
			++LB;
		if (isupper(a) && isupper(b))
			++N;
		}
	if (N == 0)
		return;

	ScoreMx Loss = g_Loss;
	SCORE d = (SCORE) (Scale/N);
	for (unsigned i = 0; i < GLA; ++i)
		for (unsigned j = 0; j < GLB; ++j)
			g_Loss[i][j] = -d;

	int i = 0;
	int j = 0;
	for (int p = 0; p < L; ++p)
		{
		char a = A[p];
		char b = B[p];
		if (isupper(a) && isupper(b))
			Loss[i][j] += d;
		if (!IsGapChar(a))
			++i;
		if (!IsGapChar(b))
			++j;
		}
	}
