#include <stdlib.h>
#include <time.h>
#ifdef _MSC_VER
#include <process.h>
#else
#include <unistd.h>
#endif
#include "utils.h"

/***
A multiply-with-carry random number generator, see:
http://en.wikipedia.org/wiki/Multiply-with-carry

The particular multipliers used here were found on
the web where they are attributed to George Marsaglia.
***/

static uint32 g_X[5];

void WriteRandState(FILE *f)
	{
	WriteStdioFile(f, g_X, sizeof(g_X));
	}

void ReadRandState(FILE *f)
	{
	ReadStdioFile(f, g_X, sizeof(g_X));
	}

void LogRandState()
	{
	Log("\n");
	Log("Rand state: ");
	for (int i = 0; i < 5; ++i)
		Log(" %08X", g_X[i]);
	Log("\n");
	}

uint32 RandInt32()
	{
	uint64 Sum = 2111111111*(uint64) g_X[3] + 1492*(uint64) g_X[2] +
	  1776*(uint64) g_X[1] + 5115*(uint64) g_X[0] + g_X[4];
	g_X[3] = g_X[2];
	g_X[2] = g_X[1];
	g_X[1] = g_X[0];
	g_X[4] = (uint32) (Sum >> 32);
	g_X[0] = (uint32) Sum;
	return g_X[0];
	}

void InitRand(uint32 Seed)
	{
	Log("Seed = %u\n", Seed);
	srand(Seed);

// rand() on both Win32 and Linux gcc had horrible spectral properties,
// but good enough to spin us up...
	for (unsigned i = 0; i < 5; i++)
		g_X[i] = rand();

	for (unsigned i = 0; i < 100; i++)
		RandInt32();
	}

static bool Init()
	{
	unsigned Seed = (unsigned) (time(0)*getpid());
	InitRand(Seed);
	return true;
	}
static bool InitDone = Init();
