#include "mus.h"

#define TRACEDP		0

static bool g_LogTB = false;

const byte TRACEBITS_DM = 0x01;
const byte TRACEBITS_IM = 0x02;
const byte TRACEBITS_DD = 0x04;
const byte TRACEBITS_II = 0x08;

float **g_DPM;
float **g_DPD;
float **g_DPI;

char **g_TBM;
char **g_TBD;
char **g_TBI;

float GapOpen;
float GapExt;

float TermGapOpenRightA;
float TermGapOpenRightB;
float TermGapOpenLeftA;
float TermGapOpenLeftB;

float TermGapExtRightA;
float TermGapExtRightB;
float TermGapExtLeftA;
float TermGapExtLeftB;

static unsigned g_MaxL;
float *g_DPRow1;
float *g_DPRow2;
float *g_DPRow3;

byte **g_TBBit;

float **g_SubstMx;

static void TraceBackBit(unsigned LA, unsigned LB, char State, string &Path)
	{
	Path.clear();

	byte **TB = g_TBBit;

	if (g_LogTB)
		{
		Log("\n");
		Log("TraceBackBit(LA=%u, LB=%u, State=%c)\n", LA, LB, State);
		}

	size_t i = LA;
	size_t j = LB;
	for (;;)
		{
		if (g_LogTB)
			Log("i=%3d  j=%3d  state=%c\n", (int) i, (int) j, State);
		if (i == 0 && j == 0)
			break;
		Path.push_back(State);

		byte t;
		switch (State)
			{
		case 'M':
			asserta(i > 0 && j > 0);
			t = TB[i-1][j-1];
			if (t & TRACEBITS_DM)
				State = 'D';
			else if (t & TRACEBITS_IM)
				State = 'I';
			else
				State = 'M';
			--i;
			--j;
			break;

		case 'D':
			asserta(i > 0);
			t = TB[i-1][j];
			if (t & TRACEBITS_DD)
				State = 'D';
			else
				State = 'M';
			--i;
			break;

		case 'I':
			asserta(j > 0);
			t = TB[i][j-1];
			if (t & TRACEBITS_II)
				State = 'I';
			else
				State = 'M';
			--j;
			break;

		default:
			Die("TraceBackBit, invalid state %c", State);
			}
		}

	reverse(Path.begin(), Path.end());
	}

float NWFast(const vector<byte> &A, const vector<byte> &B,
  const ParamVec &Params, bool DoLoss, string &Path)
	{
	const unsigned LA = SIZE(A);
	const unsigned LB = SIZE(B);

	const ScoreMx &Mxx = GetSubstMx();
	g_SubstMx = Mxx;

	unsigned MaxL = max(LA, LB);
	if (MaxL > g_MaxL)
		{
		if (g_DPRow1 != 0)
			{
			delete[] g_DPRow1;
			delete[] g_DPRow2;
			delete[] g_DPRow3;
			}
		for (unsigned i = 0; i < g_MaxL; ++i)
			delete[] g_TBBit[i];

		g_MaxL = MaxL + 128;
		g_DPRow1 = new float[g_MaxL+1];
		g_DPRow2 = new float[g_MaxL+1];
		g_DPRow3 = new float[g_MaxL+1];

		g_TBBit = new byte*[g_MaxL+1];
		for (unsigned i = 0; i < g_MaxL; ++i)
			g_TBBit[i] = new byte[g_MaxL+1];
		}

	GapOpen = (float) Params.Get(P_g);
	GapExt = (float) Params.Get(P_e);

	float tg = (float) Params.Get(P_tg);
	float te = (float) Params.Get(P_te);
	float mf = (float) Params.Get(P_mf);
#if	TRACEDP
	Log("g=%g e=%g tg=%g te=%g mf=%g\n", GapOpen, GapExt, tg, te, mf);
#endif
	TermGapOpenRightA = tg;
	TermGapOpenRightB = tg;
	TermGapOpenLeftA = tg;
	TermGapOpenLeftB = tg;

	TermGapExtRightA = te;
	TermGapExtRightB = te;
	TermGapExtLeftA = te;
	TermGapExtLeftB = te;

// Set params
	float **Mx = g_SubstMx;
	
	float *Mprev = g_DPRow1;
	float *Mcurr = g_DPRow2;
	float *Drow = g_DPRow3;
	//float **DPM = g_DPM;
	//float **DPD = g_DPD;
	//float **DPI = g_DPI;
	
	byte **TB = g_TBBit;
	
// Init origin
	Mprev[0] = float (0) /*SM*/;
	Mcurr[0] = MINUS_INFINITY;
	Drow[0] = MINUS_INFINITY;
	
	for (unsigned j = 1; j <= LB; ++j)
		{
		Mprev[j] = MINUS_INFINITY;
		Drow[j] = MINUS_INFINITY;
		}
	
// Main loop
	float OpenA = TermGapOpenLeftA;
	float ExtA = TermGapExtLeftA;
	float Iprev;
	for (unsigned i = 0; i < LA; ++i)
		{
#if TRACEDP
		{
		Log("\n");
		Log("i=%u\n", i);
		Log("Irow =");
		}
#endif
		const byte a = (byte) A[i];
		const float *MxRow = Mx[a];
		float OpenB = TermGapOpenLeftB;
		float ExtB = TermGapExtLeftB;
		Iprev = MINUS_INFINITY;
		
		byte *TBrow = TB[i];
		for (unsigned j = 0; j < LB; ++j)
			{
			byte TraceBits = 0;
			
		// xM
			{
			float mm = Mprev[j];
			float dm = Drow[j];
			if (mm >= dm && mm >= Iprev)
				Mcurr[j+1] = mm;
			else if (dm >= mm && dm >= Iprev)
				{
				Mcurr[j+1] = dm;
				TraceBits |= TRACEBITS_DM;
				}
			else
				{
				Mcurr[j+1] = Iprev;
				TraceBits |= TRACEBITS_IM;
				}
			const byte b = (byte) B[j];
			float Match = MxRow[b]*mf;
			if (DoLoss)
				Match += g_Loss[i][j];
			Mcurr[j+1] += Match;
			}
			
		// xD
			{
			float md = Mprev[j] + OpenB;
			float dd = Drow[j] + ExtB;
			if (md >= dd)
				Drow[j] = md;
			else
				{
				Drow[j] = dd;
				TraceBits |= TRACEBITS_DD;
				}
			}
			
		// xI
			{
			float mi = Mprev[j] + OpenA;
			float ii = Iprev + ExtA;
			if (mi >= ii)
				Iprev = mi;
			else
				{
				Iprev = ii;
				TraceBits |= TRACEBITS_II;
				}
#if	TRACEDP
			Log(" %8.3g", Iprev);
#endif
			OpenB = GapOpen;
			ExtB = GapExt;
			}
			
			TBrow[j] = TraceBits;
			}
		
	// Special case for end of Drow[]
		float md = Mprev[LB] + TermGapOpenRightB;
		float dd = Drow[LB] + TermGapExtRightB;
		if (md >= dd)
			{
			Drow[LB] = md;
			TBrow[LB] = 0;
			}
		else
			{
			Drow[LB] = dd;
			TBrow[LB] = TRACEBITS_DD;
			}
#if TRACEDP
		{
		Log("\nDrow =");
		for (unsigned j = 0; j < LB; ++j)
			Log(" %8.3g", Drow[j]);
		Log("\nMcurr=");
		for (unsigned j = 0; j < LB; ++j)
			Log(" %8.3g", Mcurr[j]);
		Log("\n");
		}
#endif
		swap(Mcurr, Mprev);
		Mprev[0] = MINUS_INFINITY;
		
		OpenA = GapOpen;
		ExtA = GapExt;
		}

#if TRACEDP
		Log("\nLastI=");
#endif
	
// Special case for last row of DPI
	Iprev = MINUS_INFINITY;
	byte *TBrow = TB[LA];
	for (unsigned j = 0; j < LB; ++j)
		{
		float mi = Mprev[j] + TermGapOpenRightA;
		float ii = Iprev + TermGapExtRightA;
		if (mi >= ii)
			{
			Iprev = mi;
			TBrow[j] = 0;
			}
		else
			{
			Iprev = ii;
			TBrow[j] = TRACEBITS_II;
			}
#if	TRACEDP
		Log(" %8.3g", Iprev);
#endif
		}
#if	TRACEDP
	Log("\n");
#endif

	float Me = Mprev[LB];
	float De = Drow[LB];
	float Ie = Iprev;

	char State;
	float Score;
	if (Me >= De && Me >= Ie)
		{
		State = 'M';
		Score = Me;
		}
	else if (De >= Me && De >= Ie)
		{
		State = 'D';
		Score = De;
		}
	else
		{
		State = 'I';
		Score = Ie;
		}

	TraceBackBit(LA, LB, State, Path);
	return Score;
	}
