#include "mus.h"

vector<bool> ParamVec::m_Active;
bool ParamVec::m_FixMF = false;

double ParamVec::m_Alphas[AlnParamCount];

bool ParamVec::InitAlphas()
	{
	for (unsigned i = 0; i < AlnParamCount; ++i)
		m_Alphas[i] = 1.0;

	m_Alphas[P_g] = 5.0;
	m_Alphas[P_dg] = 5.0;
	m_Alphas[P_mg] = 5.0;

	m_Alphas[P_e] = 0.3;
	m_Alphas[P_de] = 0.3;
	m_Alphas[P_me] = 0.3;

	return true;
	}
static bool g_InitAlphasDone = ParamVec::InitAlphas();

bool ParamVec::IsNegativeParam(uns ParamIndex)
	{
	switch (ParamIndex)
		{
	case P_mf:
	case P_dx:
	case P_h:
	case P_b:
	case P_f:
		return false;

	default:
		break;
		}
	return true;
	}

void ParamVec::FixMF(bool Fix)
	{
	m_FixMF = Fix;
	m_Active[P_mf] = !Fix;
	}

void ParamVec::Clear()
	{
	m_Vec.clear();
	}

ParamVec::ParamVec()
	{
	m_Vec.resize(AlnParamCount, 0);
	}

void ParamVec::Zero(double mf)
	{
	m_Vec.clear();
	m_Vec.resize(AlnParamCount, 0);
	Set(P_mf, mf);
	}

size_t ParamVec::size() const
	{
	return m_Vec.size();
	}

void ParamVec::Set(uns Param, double Value)
	{
	asserta(Param < AlnParamCount);
	m_Vec[Param] = (float) Value;
	}

double ParamVec::Get(uns Param) const
	{
	asserta(Param < AlnParamCount);
	return m_Vec[Param];
	}

void ParamVec::SetActive(uns Param, bool On)
	{
	asserta(Param < AlnParamCount);
	m_Active[Param] = On;
	}

bool ParamVec::IsActive(uns Param)
	{
	asserta(Param < AlnParamCount);
	return m_Active[Param];
	}

bool ParamVec::IsActiveOrMF(uns Param)
	{
	asserta(Param < AlnParamCount);
	return Param == P_mf || m_Active[Param];
	}

uns ParamVec::GetActiveParamCount()
	{
	uns Count = 0;
	for (uns j = 0; j < AlnParamCount; ++j)
		if (m_Active[j])
			++Count;
	return Count;
	}

uns ParamVec::IndexToActiveIndex(uns ParamIndex)
	{
	if (!m_Active[ParamIndex])
		Die("IndexToActiveIndex(%u), not active", ParamIndex);
	uns Count = 0;
	for (uns j = 0; j < ParamIndex; ++j)
		if (m_Active[j])
			++Count;
	return Count;
	}

uns ParamVec::ActiveIndexToIndex(uns ActiveParamIndex)
	{
	uns Count = 0;
	for (uns j = 0; j < AlnParamCount; ++j)
		{
		if (m_Active[j])
			{
			if (Count == ActiveParamIndex)
				return j;
			++Count;
			}
		}
	Die("ActiveIndexToIndex(%u), %u active", ActiveParamIndex, Count);
	UNREACHABLE_RETURN(0);
	}

double ParamVec::GetActiveParam(uns ActiveParamIndex) const
	{
	uns ParamIndex = ActiveIndexToIndex(ActiveParamIndex);
	return Get(ParamIndex);
	}

void ParamVec::FromCmdLine()
	{
	const string Model = RequiredValueOpt("model");

	double g = 0;
	double e = 0;
	double tg = 0;
	double te = 0;
	double mg = 0;
	double me = 0;
	double dx = 0;
	double de = 0;
	double mf = 1;
	double mp = 1;
	double h = 0;
	double b = 0;
	double f = 0;

	if (Model == "g1" || Model == "L1")
		{
		e = RequiredFloatOpt("e");
		g = e;
		te = e;
		tg = e;

		SetActive(P_e);
		}
	else if (Model == "g2")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		tg = g;
		te = e;

		SetActive(P_g);
		SetActive(P_e);
		}
	else if (Model == "g2c")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		tg = g;
		te = e;

		SetActive(P_g);
		SetActive(P_e);
		}
	else if (Model == "g2f")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		f  = RequiredFloatOpt("f");

		tg = g;
		te = e;

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_f);
		}
	else if (Model == "g3")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		tg = RequiredFloatOpt("tg");
		te = e;

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_tg);
		}
	else if (Model == "e3")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		te = RequiredFloatOpt("te");
		tg = g;

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_te);
		}
	else if (Model == "g4")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		tg = RequiredFloatOpt("tg");
		te = RequiredFloatOpt("te");

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_tg);
		SetActive(P_te);
		}
	else if (Model == "g6")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		tg = RequiredFloatOpt("tg");
		te = RequiredFloatOpt("te");
		mg = RequiredFloatOpt("mg");
		me = RequiredFloatOpt("me");

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_tg);
		SetActive(P_te);
		SetActive(P_mg);
		SetActive(P_me);
		}
	else if (Model == "da")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		de = RequiredFloatOpt("de");
		tg = g;
		te = e;
		if (ValueOpt("dx"))
			dx = RequiredFloatOpt("dx");
		else
			{
			double dg = RequiredFloatOpt("dg");
			if (fabs(e - de) < 1e-6)
				dx = 1e6;
			else
				{
				dx = (dg - g)/(e - de);
				dx /= 10;
				}
			}

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_dx);
		SetActive(P_de);
		}

	else if (Model == "da2")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");

		SetActive(P_g);
		SetActive(P_e);
		}

	else if (Model == "d3")
		{
		g = RequiredFloatOpt("g");
		e = RequiredFloatOpt("e");
		de = 0;
		if (ValueOpt("dx"))
			dx = RequiredFloatOpt("dx");
		else
			{
			double dg = RequiredFloatOpt("dg");
			if (fabs(e - de) < 1e-6)
				dx = 1e6;
			else
				{
				dx = (dg - g)/(e - de);
				dx /= 10;
				}
			}

		SetActive(P_g);
		SetActive(P_e);
		SetActive(P_dx);
		}
	else
		Die("-model required");

	FloatOpt("mf", &mf);
	FloatOpt("mp", &mp);
	FloatOpt("h", &h);
	FloatOpt("b", &b);

	Set(P_g, -g);
	Set(P_tg, -tg);
	Set(P_mg, -mg);
	Set(P_e, -e);
	Set(P_te, -te);
	Set(P_me, -me);
	Set(P_dx, dx);
	Set(P_de, -de);
	Set(P_mf, mf);
	Set(P_h, h);
	Set(P_b, b);
	Set(P_f, f);

	if (ValueOpt("h") != 0)
		ParamVec::SetActive(P_h);
	if (ValueOpt("b") != 0)
		ParamVec::SetActive(P_b);
	}

string &ParamVec::ToStr(string &Str) const
	{
	Str.clear();
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		char tmp[64];
		double v = m_Vec[i];
		if (i != P_mf && m_Vec[P_mf] != 0)
			v /= m_Vec[P_mf];
		if (IsNegativeParam(i))
			v = -v;
		sprintf(tmp, "%s%s=%.5g", (i == 0 ? "" : " "), AlnParamToStr(i), v);
		Str += tmp;
		}
	if (IsActive(P_dx))
		{
		char tmp[64];
		sprintf(tmp, " DAL=%u dg=%.5g", GetDAL(), -Get(P_dg));
		Str += tmp;
		}
	return Str;
	}

void ParamVec::LogAll() const
	{
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		double v = m_Vec[i];
		Log("%s%s%s=%.5g",
		  (i == 0 ? "" : " "),
		  (m_Active[i] ? "" : "~"),
		  AlnParamToStr(i), v);
		}
	}

void ParamVec::LogMe(bool NL, bool Scale) const
	{
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		bool Active = m_Active[i];
		if (i == P_dg && IsActive(P_dx))
			Active = true;
		if (!Active)
			continue;
		double v = m_Vec[i];
		if (Scale && i != P_mf && m_Vec[P_mf] != 0)
			v /= m_Vec[P_mf];
		Log("%s%s=%.5g", (i == 0 ? "" : " "), AlnParamToStr(i), v);
		}
	if (NL)
		Log("\n");
	}

string &ParamVec::ToCmdLine(string &Opts) const
	{
	Opts.clear();
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		if (i == P_mf)
			continue;
		double v = m_Vec[i];
		double s = m_Vec[P_mf];
		if (s == 0)
			s = 1;
		v /= m_Vec[P_mf];

		char Tmp[64];
		sprintf(Tmp, "-%s %.5g", AlnParamToStr(i), -v);
		if (!Opts.empty())
			Opts += " ";
		Opts += Tmp;
		}
	return Opts;
	}

string &ParamVec::ToSepStr(string &Str, char Sep) const
	{
	Str.clear();
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		if (i == P_mf)
			continue;
		double v = m_Vec[i];
		double s = m_Vec[P_mf];
		if (s == 0)
			s = 1;
		v /= m_Vec[P_mf];

		if (IsNegativeParam(i))
			v = -v;

		char Tmp[64];
		sprintf(Tmp, "%s=%.5g%c", AlnParamToStr(i), v, Sep);
		Str += Tmp;
		}
	return Str;
	}

void ParamVec::LogRow(bool DoScale) const
	{
	double Scale = (double) (DoScale ? Get(P_mf) : 1.0);
	if (Scale == 0.0)
		Scale = 1.0;

	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		double v = Get(i);
		if (DoScale && i != P_mf)
			v /= Scale;

		if (fabs(v) >= 0.001 && fabs(v) <= 100)
			Log("  %9.4f", v);
		else
			Log("  %9.2g", v);
		}
	}

void ParamVec::LogRowHdr1(char c)
	{
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		if (c == 0)
			Log("  %9.9s", AlnParamToStr(i));
		else
			Log("  %8.8s%c", AlnParamToStr(i), c);
		}
	}

void ParamVec::LogRowHdr2()
	{
	for (uns i = 0; i < AlnParamCount; ++i)
		if (m_Active[i])
			Log("  ---------");
	}

bool ParamVec::operator<(const ParamVec &rhs) const
	{
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		double p = Get(i);
		double rp = rhs.Get(i);
		if (p != rp)
			return p < rp;
		}
	return false;
	}

void ParamVec::Div(double x, bool Do_mf)
	{
	asserta(x != 0);
	for (uns i = 0; i < AlnParamCount; ++i)
		{
		if (!m_Active[i])
			continue;
		if (i == P_mf && !Do_mf)
			continue;
		m_Vec[i] /= (float) x;
		}
	}

uns ParamVec::GetDAL() const
	{
	if (!IsActive(P_dx))
		return 0;

	return (uns) (10*Get(P_dx));
	}
