#ifndef profile_h
#define profile_h

#include "letters.h"
#include "dimer.h"

char LetterToChar(byte Letter);
uns GetAlphabetSize();

class ParamVec;

struct ProfColData
	{
	ProfColData(const ProfColData &rhs)
		{
		Die("ProfColData cpy c'tor not properly implemented!");
		Occ = 1.0;
		Chars = rhs.Chars;
		Letters = rhs.Letters;
		LetterFreqs = rhs.LetterFreqs;
		DimerFreqs = new double[DimerCount];
		memcpy(DimerFreqs, rhs.DimerFreqs, DimerCount*sizeof(double));
		}

	ProfColData &operator=(const ProfColData &rhs)
		{
		const uns AlphabetSize = GetAlphabetSize();
		Chars = rhs.Chars;
		memcpy(Letters, rhs.Letters, AlphabetSize*sizeof(byte));
		memcpy(LetterFreqs, rhs.LetterFreqs, AlphabetSize*sizeof(double));
		memcpy(DimerFreqs, rhs.DimerFreqs, DimerCount*sizeof(double));
		memcpy(SubstScores, rhs.SubstScores, AlphabetSize*sizeof(SCORE));

#define	m(x)	memcpy(DimerSubstScores##x, rhs.DimerSubstScores##x, DimerCount*sizeof(SCORE));

		m(MM)

		m(SD)
		m(MD)
		m(MDt)
		m(DM)
		m(DMt)
		m(DD)
		m(DDt)

		m(SX)
		m(MX)
		m(MXt)
		m(XM)
		m(XMt)
		m(XX)
		m(XXt)

#undef m
		GapOpen = rhs.GapOpen;
		GapClose = rhs.GapClose;
		Hydro = rhs.Hydro;
		Occ = rhs.Occ;
		LetterCount = rhs.LetterCount;
		return *this;
		}

	ProfColData()
		{
		extern uns GetAlphabetSize();
		uns AlphabetSize = GetAlphabetSize();

		DimerFreqs = new double[DimerCount];
		Letters = new byte[AlphabetSize];
		LetterFreqs = new double[AlphabetSize];
		SubstScores = new SCORE[AlphabetSize];
		DimerSubstScoresMM = new SCORE[DimerCount];

		DimerSubstScoresSD = new SCORE[DimerCount];
		DimerSubstScoresMD = new SCORE[DimerCount];
		DimerSubstScoresMDt = new SCORE[DimerCount];
		DimerSubstScoresDM = new SCORE[DimerCount];
		DimerSubstScoresDMt = new SCORE[DimerCount];
		DimerSubstScoresDD = new SCORE[DimerCount];
		DimerSubstScoresDDt = new SCORE[DimerCount];

		DimerSubstScoresSX = new SCORE[DimerCount];
		DimerSubstScoresMX = new SCORE[DimerCount];
		DimerSubstScoresMXt = new SCORE[DimerCount];
		DimerSubstScoresXM = new SCORE[DimerCount];
		DimerSubstScoresXMt = new SCORE[DimerCount];
		DimerSubstScoresXX = new SCORE[DimerCount];
		DimerSubstScoresXXt = new SCORE[DimerCount];

		Clear();
		}

	~ProfColData()
		{
		Chars.clear();
		delete[] Letters;
		delete[] LetterFreqs;
		delete[] DimerFreqs;
		delete[] SubstScores;
		delete[] DimerSubstScoresMM;

		delete[] DimerSubstScoresSD;
		delete[] DimerSubstScoresMD;
		delete[] DimerSubstScoresMDt;
		delete[] DimerSubstScoresDM;
		delete[] DimerSubstScoresDMt;
		delete[] DimerSubstScoresDD;
		delete[] DimerSubstScoresDDt;

		delete[] DimerSubstScoresSX;
		delete[] DimerSubstScoresMX;
		delete[] DimerSubstScoresMXt;
		delete[] DimerSubstScoresXM;
		delete[] DimerSubstScoresXMt;
		delete[] DimerSubstScoresXX;
		delete[] DimerSubstScoresXXt;
		}

	void Clear()
		{
		extern uns GetAlphabetSize();
		uns AlphabetSize = GetAlphabetSize();
		for (uns i = 0; i < AlphabetSize; ++i)
			SubstScores[i] = 0;

		for (uns i = 0; i < DimerCount; ++i)
			{
			DimerFreqs[i] = 0.0;
			DimerSubstScoresMM[i] = 0.0f;

			DimerSubstScoresSD[i] = 0.0f;
			DimerSubstScoresMD[i] = 0.0f;
			DimerSubstScoresMDt[i] = 0.0f;
			DimerSubstScoresDM[i] = 0.0f;
			DimerSubstScoresDMt[i] = 0.0f;
			DimerSubstScoresDD[i] = 0.0f;
			DimerSubstScoresDDt[i] = 0.0f;

			DimerSubstScoresSX[i] = 0.0f;
			DimerSubstScoresMX[i] = 0.0f;
			DimerSubstScoresMXt[i] = 0.0f;
			DimerSubstScoresXM[i] = 0.0f;
			DimerSubstScoresXMt[i] = 0.0f;
			DimerSubstScoresXX[i] = 0.0f;
			DimerSubstScoresXXt[i] = 0.0f;
			}

		GapOpen = 0;
		GapClose = 0;
		Hydro = 0;
		LetterCount = 0;
		}

	void LogMe() const
		{
		Log("Chars:  %s\n", Chars.c_str());

		Log("Freqs:  ");
		for (uns i = 0; i < LetterCount; ++i)
			Log(" %c=%5.3f", LetterToChar(Letters[i]), LetterFreqs[i]);		
		for (uns i = 0; i < DimerCount; ++i)
			Log(" %s=%.3g", IntDimerToStr(i), DimerFreqs[i]);
		Log("\n");

		uns GetAlphabetSize();
		uns AlphabetSize = GetAlphabetSize();
		Log("Subst:  ");
		for (uns i = 0; i < AlphabetSize; ++i)
			Log(" %c=%g", LetterToChar((byte) i), SubstScores[i]);
		Log("\n");
		}

	string Chars;
	byte *Letters;
	double *LetterFreqs;
	SCORE *SubstScores;

	SCORE *DimerSubstScoresMM;
	SCORE *DimerSubstScoresSD;
	SCORE *DimerSubstScoresMD;
	SCORE *DimerSubstScoresMDt;
	SCORE *DimerSubstScoresDM;
	SCORE *DimerSubstScoresDMt;
	SCORE *DimerSubstScoresDD;
	SCORE *DimerSubstScoresDDt;

	SCORE *DimerSubstScoresSX;
	SCORE *DimerSubstScoresMX;
	SCORE *DimerSubstScoresMXt;
	SCORE *DimerSubstScoresXM;
	SCORE *DimerSubstScoresXMt;
	SCORE *DimerSubstScoresXX;
	SCORE *DimerSubstScoresXXt;

	double *DimerFreqs;
	SCORE GapOpen;
	SCORE GapClose;
	double Hydro;
	double Occ;
	uns LetterCount;
	};

class Profile
	{
public:
	Profile()
		{
		m_ColCount = 0;
		}

	~Profile()
		{
		Clear();
		}

	void Clear()
		{
		if (m_ColCount > 0)
			delete[] m_Cols;
		m_ColCount = 0;
		}
	uns GetColCount() const
		{
		return m_ColCount;
		}
	uns GetSeqCount() const
		{
		if (GetColCount() == 0)
			return 0;
		return SIZE(m_Cols[0].Chars);
		}
	ProfColData &GetCol(uns ColIndex)
		{
		assert(ColIndex < GetColCount());
		return m_Cols[ColIndex];
		}
	const ProfColData &GetCol(uns ColIndex) const
		{
		assert(ColIndex < GetColCount());
		return m_Cols[ColIndex];
		}

	void LogMe() const;
	void FromMSA(const vector<string > &MSA, const vector<string> &Labels,
	  const vector<double> &SeqWeights, const ParamVec &Params);
	double GetHydroCol(unsigned ColIndex) const;
	void SetSubstScores(ProfColData &Col);
	void SetDimerScores(const double *DimerFreqs, const ScoreMx &Mx,
	  SCORE *Scores);
	void SetScores(const ParamVec &Params, double SumSeqWeights);
	void SetHydro(const vector<string> &MSA, unsigned WindowLength, bool Dropoff);
	void SetMAFFTGapPenalties(const ParamVec &Params);
	void GetColRange(uns ColIndex, uns ColCount, Profile &P) const;

private:
	ProfColData *m_Cols;
	uns m_ColCount;
	};

void PathFromMSAPair(const vector<string> &MSA1, const vector<string> &MSA2,
  string &Path);
SCORE PPScoreGapRecurse(
  const vector<string> &MSA1, const vector<string> &Labels1, const vector<double> &SeqWeights1,
  const vector<string> &MSA2, const vector<string> &Labels2, const vector<double> &SeqWeights2,
  const ParamVec &Params);

#endif // profile_h
