#include "mus.h"

float NWFast(const vector<byte> &A, const vector<byte> &B,
  const ParamVec &Params, bool DoLoss, string &Path);

static PWA2 g_PWA2;
static PWA4 g_PWA4;

PairWiseAligner *g_PWA = 0;

void PairWiseAligner::InitFromCmdLine()
	{
	ParamVec::m_Active.resize(AlnParamCount);
	const string Model = RequiredValueOpt("model");
	if (Model == "g2")
		g_PWA = &g_PWA2;
	else if (Model == "g4")
		g_PWA = &g_PWA4;
	else
		Die("Invalid -model %s", Model.c_str());

	g_PWA->SetActiveParams();
	g_PWA->SetOpts();
	Log("PWA=%s\n", g_PWA->GetName());
	}

void PairWiseAligner::InitFromCmdLine(ParamVec &Params)
	{
	InitFromCmdLine();
	g_PWA->SubclassInitParamsFromCmdLine(Params);
	g_PWA->SubclassSetParams(Params);
	g_PWA->SetActiveParams();
	}

void PairWiseAligner::SetParams(ParamVec &Params) const
	{
	SubclassSetParams(Params);
	}

/////////////////////////////////////////////////////////////////////////

SCORE PWA4::NW(const vector<byte> &A, const vector<byte> &B,
  const ParamVec &Params, bool DoLoss, 
  bool /*ALeftTrunc*/, bool /*ARightTrunc*/, bool /*BLeftTrunc*/, bool /*BRightTrunc*/,
  string &Path) const
	{
	ParamVec Params2(Params);
	SubclassSetParams(Params2);
	return ::NWFast(A, B, Params2, DoLoss, Path);
	}

void PWA4::SetActiveParams() const
	{
	ParamVec::SetActive(P_g);
	ParamVec::SetActive(P_e);
	ParamVec::SetActive(P_tg);
	ParamVec::SetActive(P_te);
	ParamVec::SetActive(P_mf);
	}

void PWA4::SubclassInitParamsFromCmdLine(ParamVec &Params) const
	{
	Params.FromCmdLine();
	}

void PWA4::SubclassSetParams(ParamVec &/*Params*/) const
	{
	}

/////////////////////////////////////////////////////////////////////////

SCORE PWA2::NW(const vector<byte> &A, const vector<byte> &B,
  const ParamVec &Params, bool DoLoss,
  bool /*ALeftTrunc*/, bool /*ARightTrunc*/, bool /*BLeftTrunc*/, bool /*BRightTrunc*/,
  string &Path) const
	{
	ParamVec Params2(Params);
	SubclassSetParams(Params2);
	return ::NWFast(A, B, Params2, DoLoss, Path);
	}

void PWA2::SubclassInitParamsFromCmdLine(ParamVec &Params) const
	{
	Params.FromCmdLine();
	}

void PWA2::SubclassSetParams(ParamVec &Params) const
	{
	SCORE g = (SCORE) Params.Get(P_g);
	SCORE e = (SCORE) Params.Get(P_e);
	switch (m_TermGapStyle)
		{
	case TGS_Free:
		Params.Set(P_tg, 0);
		Params.Set(P_te, 0);
		break;

	case TGS_ExtOnly:
		Params.Set(P_tg, e);
		Params.Set(P_te, e);
		break;

	case TGS_Half:
		Params.Set(P_tg, g/2);
		Params.Set(P_te, e/2);
		break;

	case TGS_Full:
		Params.Set(P_tg, g);
		Params.Set(P_te, e);
		break;

	default:
		asserta(false);
		}
	}

void PWA2::SetActiveParams() const
	{
	ParamVec::SetActive(P_g);
	ParamVec::SetActive(P_e);
	ParamVec::SetActive(P_mf);
	}

void PWA2::SetOpts()
	{
	const char *TG = ValueOpt("termgaps");
	if (TG == 0)
		TG = "full";

	if (!strcmp(TG, "free"))
		m_TermGapStyle = TGS_Free;
	else if (!strcmp(TG, "ext"))
		m_TermGapStyle = TGS_ExtOnly;
	else if (!strcmp(TG, "half"))
		m_TermGapStyle = TGS_Half;
	else if (!strcmp(TG, "full"))
		m_TermGapStyle = TGS_Full;
	else
		Die("Invalid -termgaps %s", TG);
	Log("PWA2::TermGapStyle %s\n", TG);
	}
