#include "mus.h"

bool IsFirstState(char c)
	{
	switch (c)
		{
	case 'M':
	case 'D':
	case 'X':
	case S_MA:
	case S_MB:
	case S_MC:
	case S_DA:
	case S_DC:
	case S_UA:
	case S_UM:
		return true;
	default:
		return false;
		}
	}

bool IsSecondState(char c)
	{
	switch (c)
		{
	case 'M':
	case 'I':
	case 'Y':
	case S_MA:
	case S_MB:
	case S_MC:
	case S_IA:
	case S_IC:
	case S_UB:
	case S_UM:
		return true;
	default:
		return false;
		}
	}

bool IsAlignedState(char c)
	{
	switch (c)
		{
	case S_UM:
	case S_UA:
	case S_UB:
		return false;
	default:
		return true;
		}
	}

void LogPairWiseSubAln(const string &A, const string &B, const string &Path)
	{
	LogPWAln(A, B, Path, true);
	}

void AlnStr(const string &A, const string &Path, bool First, string &sA,
  bool ValidateLength)
	{
	sA.clear();
	uns i = 0;
	for (string::const_iterator p = Path.begin(); p != Path.end(); ++p)
		{
		char c = *p;
		if (First)
			{
			if (IsFirstState(c))
				{
				assert(i < A.size());
				char a = A[i++];
				if (IsAlignedState(c))
					a = toupper(a);
				else
					a = tolower(a);
				sA.push_back(a);
				}
			else
				sA.push_back('-');
			}
		else
			{
			if (IsSecondState(c))
				{
				assert(i < A.size());
				char a = A[i++];
				if (IsAlignedState(c))
					a = toupper(a);
				else
					a = tolower(a);
				sA.push_back(a);
				}
			else
				sA.push_back('-');
			}
		}

	if (ValidateLength && i != (uns) A.size())
		{
		Log("\n");
		Log("Str =%s\n", A.c_str());
		Log("Path=%s\n", Path.c_str());
		Die("AlnStr, path uses %u letters < input string length %u",
		  i, (uns) A.size());
		}
	}

void LogPWAln(const string &A, const string &B, const string &Path,
  bool /*SubAln*/)
	{
	string sA;
	string sB;
	AlnStr(A, Path, true, sA);
	AlnStr(B, Path, false, sB);

	Log("\n");
	Log("%s\n", sA.c_str());
	Log("%s\n", sB.c_str());
	Log("%s\n", Path.c_str());
	}

void AlnToPath(const string &A, const string &B, string &Path)
	{
	Path.clear();
	const int L = (int) A.size();
	Path.reserve(L);
	if ((int) B.size() != L)
		Die("AlnToPath, different lengths");
	for (int p = 0; p < L; ++p)
		{
		char a = A[p];
		char b = B[p];
		bool GapA = IsGapChar(a);
		bool GapB = IsGapChar(b);
		if (GapA && GapB)
			Die("AlnToPath, gap-gap col.");
		if (GapA)
			Path.push_back('I');
		else if (GapB)
			Path.push_back('D');
		else
			Path.push_back('M');
		}
	}

uns GetFirstCount(const string &Path)
	{
	uns Count = 0;
	for (string::const_iterator p = Path.begin(); p != Path.end(); ++p)
		if (IsFirstState(*p))
			++Count;
	return Count;
	}

uns GetSecondCount(const string &Path)
	{
	uns Count = 0;
	for (string::const_iterator p = Path.begin(); p != Path.end(); ++p)
		if (IsSecondState(*p))
			++Count;
	return Count;
	}
