#include "mus.h"

double PWSet::GetWeight(uns PairIndex) const
	{
	asserta(PairIndex < SIZE(m_PairWeights));
	return m_PairWeights[PairIndex];
	}

double PWSet::ComputeQ(const ParamVec &Params)
	{
	const uns PairCount = GetPairCount();
	asserta(PairCount > 0);
	double SumQW = 0;
	double SumWeights = 0;
	string Tmp;
	Params.ToStr(Tmp);
	ProgressStart("ComputeQ %s", Tmp.c_str());
	for (uns PairIndex = 0; PairIndex < PairCount; ++PairIndex)
		{
		if (PairIndex%100 == 0)
			ProgressStep(PairIndex, PairCount);

		double Q = ComputeQPair(PairIndex, Params);
		double Weight = GetWeight(PairIndex);
		asserta(Weight > 0);
		SumWeights += Weight;
		SumQW += Weight*Q;
		}

	double Q = SumQW/SumWeights;
	SetProgressMsg("ComputeQ %s Q=%7.5f", Tmp.c_str(), Q);
	ProgressDone();
	return Q;
	}

double PWSet::ComputeQPair(uns PairIndex, const ParamVec &Params) const
	{
	RefPairData Ref;
	GetRefPair(PairIndex, Ref);

	ParamVec Params2(Params);
	g_PWA->SubclassSetParams(Params2);

	string Path;
	SCORE Score = g_PWA->NW(*Ref.UVA, *Ref.UVB, Params2, false, 
	  Ref.ALeftTrunc, Ref.ARightTrunc, Ref.BLeftTrunc, Ref.BRightTrunc, Path);

	vector<string> TestPWA(2);
	AlnStr(*Ref.USA, Path, true, TestPWA[0]);
	AlnStr(*Ref.USB, Path, false, TestPWA[1]);
	const string &TestA = TestPWA[0];
	const string &TestB = TestPWA[1];

	double Q;
	double TC;
	QScore("?", TestPWA, *Ref.Labels, *Ref.RefPWA, *Ref.Labels, Q, TC, false);

	if (g_LogAlns)
		{
		const string &Name = Ref.GetName();
		Log("\n");
		Log("%s\n", TestA.c_str());
		Log("%s\n", TestB.c_str());
		Log("%s\n", Path.c_str());
		Log("%s Q=%5.3f Score=%g\n", Name.c_str(), Q, Score);
		}

	if (!g_LogAlns && (m_Iter >= m_LogIterFrom && m_Iter <= m_LogIterTo))
		{
		Log("\n");
		Log("Q=%5.3f %s\n", Q, Ref.GetName().c_str());
		Log("Params: ");
		Params.LogMe(true, true);
		Log("TestA: %s\n", TestA.c_str());
		Log("TestA: %s\n", TestB.c_str());
		Log(" RefA: %s\n", (*Ref.RefPWA)[0].c_str());
		Log(" RefB: %s\n", (*Ref.RefPWA)[1].c_str());
		}

	return Q;
	}

double PWSet::ComputeQs(const ParamVec &Params, vector<double> &Qs) const
	{
	const uns N = GetPairCount();
	Qs.resize(N);
	ProgressStart("Compute Qs");
	double SumWeights = 0;
	double SumQWs = 0;
	for (uns i = 0; i < N; ++i)
		{
		if (i%100 == 0)
			ProgressStep(i, N);
		 double Q = ComputeQPair(i, Params);
		 Qs[i] = Q;

		double Weight = GetWeight(i);
		SumQWs += Weight*Q;
		SumWeights += Weight;
		}
	ProgressDone();
	return SumWeights == 0 ? 0 : SumQWs/SumWeights;
	}
