#include "mus.h"

void RandomizeVec(vector<uns> &Vec)
	{
	const unsigned N = SIZE(Vec);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned j = RandInt32()%(N-i);
		swap(Vec[j], Vec[N-i-1]);
		}
	}

void PWSet::FromFileM(const char *FileName, MSASet &M, uns SubsetSize)
	{
	M.FromFile(FileName);
	uns MSACount = M.GetCount();
	if (MSACount == 0)
		Die("Empty multi set in file '%.32s'");

	uns PairCount = 0;
	for (uns i = 0; i < MSACount; ++i)
		{
		const uns SeqCount = M.GetSeqCount(i);
		PairCount += (SeqCount*(SeqCount - 1))/2;
		}

	PWSet P;
	if (SubsetSize == UINT_MAX || SubsetSize >= PairCount)
		{
		M.GetAllPairs(*this);
		return;
		}

	uns MaxPairsPerMSA = 1;
	for (;;)
		{
		uns SubsetPairCount = 0;
		for (uns i = 0; i < MSACount; ++i)
			{
			const uns SeqCount = M.GetSeqCount(i);
			uns PairCount_i = (SeqCount*(SeqCount - 1))/2;
			if (PairCount_i > MaxPairsPerMSA)
				PairCount_i = MaxPairsPerMSA;
			SubsetPairCount += PairCount_i;
			}
		if (SubsetPairCount > SubsetSize)
			{
			--MaxPairsPerMSA;
			break;
			}
		++MaxPairsPerMSA;
		}

	uns ActualSubsetSize = 0;
	for (uns i = 0; i < MSACount; ++i)
		{
		const uns SeqCount = M.GetSeqCount(i);
		uns PairCount_i = (SeqCount*(SeqCount - 1))/2;
		if (PairCount_i > MaxPairsPerMSA)
			PairCount_i = MaxPairsPerMSA;
		ActualSubsetSize += PairCount_i;
		}
	Progress("MaxPairsPerMSA=%u TargetSubset=%u Actual=%u",
	  MaxPairsPerMSA, SubsetSize, ActualSubsetSize);
	asserta(MaxPairsPerMSA > 0);

	vector<uns> MSAIndexes;
	vector<double> PairWeights;
	vector<pair<uns, uns> > SeqIndexes;
	for (uns MSAIndex = 0; MSAIndex < MSACount; ++MSAIndex)
		{
		uns SeqCount = M.GetSeqCount(MSAIndex);
		uns PairCount = (SeqCount*(SeqCount - 1))/2;
		if (PairCount == 1)
			Die("Alignment has 1 seq");
		if (PairCount <= MaxPairsPerMSA)
			{
			double Weight = 1./PairCount;
			for (uns i = 0; i < SeqCount; ++i)
				{
				for (uns j = 0; j < i; ++j)
					{
					MSAIndexes.push_back(MSAIndex);
					SeqIndexes.push_back(pair<uns, uns>(i, j));
					PairWeights.push_back(Weight);
					}
				}
			}
		else
			{
			double Weight = 1./MaxPairsPerMSA;
			vector<pair<uns, uns> > Pairs;
			for (uns i = 0; i < SeqCount; ++i)
				for (uns j = 0; j < i; ++j)
					Pairs.push_back(pair<uns, uns>(i, j));

			vector<uns> Indexes;
			for (uns i = 0; i < PairCount; ++i)
				Indexes.push_back(i);
			RandomizeVec(Indexes);
			for (uns i = 0; i < MaxPairsPerMSA; ++i)
				{
				uns k = Indexes[i];
				MSAIndexes.push_back(MSAIndex);
				SeqIndexes.push_back(Pairs[k]);
				PairWeights.push_back(Weight);
				}
			}
		}
	M.GetPWSet(MSAIndexes, SeqIndexes, PairWeights, *this);
	}
