#include "mus.h"
#include "utils.h"

static ScoreMx g_SubstMx;
static ScoreMx g_SubstMx2;

const ScoreMx &GetSubstMx(uns iMx)
	{
	return iMx == 2 ? g_SubstMx2 : g_SubstMx;
	}

static void AllocSubstMx(uns iMx)
	{
	ScoreMx &Mx = (iMx == 2 ? g_SubstMx2 : g_SubstMx);
	if (Mx != 0)
		return;
	uns N = GetAlphabetSize();
	AllocScoreMx(N, N, Mx);
	}

static void ZeroSubstMx(uns iMx)
	{
	const ScoreMx &Mx = GetSubstMx(iMx);
	uns AS = GetAlphabetSize();
	for (uns i = 0; i < AS; ++i)
		for (uns j = 0; j < AS; ++j)
			Mx[i][j] = 0;
	}

void SetScoreMx(ScoreMx &Mx, uns LA, uns LB, SCORE v)
	{
	for (uns i = 0; i < LA; ++i)
		for (uns j = 0; j < LB; ++j)
			Mx[i][j] = v;
	}

void ZeroScoreMx(ScoreMx &Mx, uns LA, uns LB)
	{
	SetScoreMx(Mx, LA, LB, 0);
	}

void AllocScoreMx(unsigned n, unsigned m, ScoreMx &Mx)
	{
	Mx = (SCORE **) malloc(n*sizeof(SCORE *));
	for (unsigned i = 0; i < n; ++i)
		Mx[i] = (SCORE *) malloc(m*sizeof(SCORE));
	}

void FreeScoreMx(unsigned n, unsigned /*m*/, ScoreMx &Mx)
	{
	for (unsigned i = 0; i < n; ++i)
		free(Mx[i]);
	free(Mx);
	Mx = 0;
	}

double GetAvgAbsSubstMxAbsValue(uns iMx)
	{
	const ScoreMx &Mx = GetSubstMx(iMx);

	uns AS = GetAlphabetSize();
	double Sum = 0;
	for (uns i = 0; i < AS; ++i)
		for (uns j = 0; j < AS; ++j)
			Sum += fabs(Mx[i][j]);
	return Sum/400.0;
	}

void LogSubstMx(uns iMx)
	{
	const ScoreMx &Mx = GetSubstMx(iMx);

	uns AS = GetAlphabetSize();
	Log("\n");
	Log("  ");
	for (uns i = 0; i < AS; ++i)
		{
		char c = LetterToChar(i);
		Log("%7c", c);
		}
	Log("\n");

	for (uns i = 0; i < AS; ++i)
		{
		Log("%c ", LetterToChar(i));
		for (uns j = 0; j < AS; ++j)
			Log("%7.2g", Mx[i][j]);
		Log("\n");
		}
	Log("\n");
	}

void LogScoreMx(const ScoreMx &Mx, uns LA, uns LB)
	{
	Log("\n");
	Log("        ");
	for (uns i = 0; i < LB; ++i)
		Log("%7u", i);
	Log("\n");

	for (uns i = 0; i < LA; ++i)
		{
		Log("%7u ", i);
		for (uns j = 0; j < LB; ++j)
			{
			SCORE s = Mx[i][j];
			if (s < MINUS_INFINITY/2)
				Log("      *");
			else
				Log("%7.2g", s);
			}
		Log("\n");
		}
	Log("\n");
	}

void ReadSubstMx(const char *FileName, ScoreMx &Mx)
	{
	FILE *f = OpenStdioFile(FileName);

	string Line;
	for (;;)
		{
		bool EndOfFile = ReadLineStdioFile(f, Line);
		if (EndOfFile)
			Die("ReadSubstMx, end-of-file in %.32s without finding data", FileName);
		if (Line.empty() || Line[0] == '#')
			continue;
		else if (Line[0] == ' ')
			break;
		else
			Die("ReadSubstMx, file %.32s has unexpected line '%.32s'",
			  FileName, Line.c_str());
		}

	vector<string> Headings;
	ParseStrVecWS(Line, Headings);

	uns N = (uns) Headings.size();
	if (N < 20)
		Die("ReadSubstMx, %.32s has %u headings", FileName, (uns) Headings.size());

	for (uns Row = 0; Row < N; ++Row)
		{
		const string &Heading = Headings[Row];
		if (Heading.size() != 1)
			Die("ReadSubstMx(%.32s), heading '%s' not one char", FileName, Heading.c_str());
		unsigned char RowChar = (unsigned char) Heading[0];
		uns RowLetter = CharToLetter(RowChar);

		bool EndOfFile = ReadLineStdioFile(f, Line);
		if (EndOfFile)
			Die("ReadSubstMx, premature end-of-file in %.32s", FileName);

		vector<string> Values;
		ParseStrVecWS(Line, Values);
		if (Values.size() != N + 1)
			Die("ReadSubstMx(%.32s), expected %u fields, got %u",
			  FileName, N + 1, (uns) Values.size());

		for (uns Col = 0; Col < N; ++Col)
			{
			const string &Heading = Headings[Col];
			if (Heading.size() != 1)
				Die("ReadSubstMx(%.32s), heading '%s' not one char", FileName, Heading.c_str());
			unsigned char ColChar = (unsigned char) Heading[0];
			uns ColLetter = CharToLetter(ColChar);

			const string &strValue = Values[Col+1];
			SCORE Value = (SCORE) atof(strValue.c_str());
			Mx[RowLetter][ColLetter] = Value;
			}
		}
	}

void ReadSubstMx(const char *FileName, uns iMx)
	{
	AllocSubstMx(iMx);
	ScoreMx &Mx = (iMx == 2 ? g_SubstMx2 : g_SubstMx);
	ReadSubstMx(FileName, Mx);
	}

void GetSubstMxFromCmdLine(uns iMx)
	{
	AllocSubstMx(iMx);
	ZeroSubstMx(iMx);
	const char *OptName = (iMx == 2 ? "mx2" : "mx");
	const char *MxName = RequiredValueOpt(OptName);
	ReadSubstMx(MxName, iMx);
	}
