#ifndef UTILS_H
#define UTILS_H

#ifdef _MSC_VER
#pragma warning(disable: 4996)	// deprecated functions
#define _CRT_SECURE_NO_DEPRECATE	1
#endif

#include <stdarg.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "myassert.h"
#include "wrapstdio.h"

#if defined(_DEBUG) && !defined(DEBUG)
#define DEBUG	1
#endif

#if defined(DEBUG) && !defined(_DEBUG)
#define _DEBUG	1
#endif

#ifndef NDEBUG
#define	DEBUG	1
#define	_DEBUG	1
#endif


#if	defined(_MSC_VER) && (!defined(DEBUG) || DEBUG == 0)
#define UNREACHABLE_RETURN(x)	/*empty*/
#else
#define UNREACHABLE_RETURN(x)	return x;
#endif

/* Cool hacks to get a string of __FILE__ ":" __LINE__ :) */
#define _level_2_str(s) _level_1_str(s)
#define _level_1_str(s) #s
#define __LINESTR__ _level_2_str(__LINE__)
#define __FL__ __FILE__ ":" __LINESTR__

#ifdef _MSC_VER
/* Unfortunately Microsoft doesn't support __VA_ARGS__ */
#define Die PrintAndDie
#else
#define Die(format, ...) PrintAndDie(__FL__ ": " format "\n" , ## __VA_ARGS__ )
#endif

#if	defined(_GNU_SOURCE) && (defined(DEBUG) || defined(_DEBUG))
#define _GLIBCXX_DEBUG
#endif

#ifdef _MSC_VER
// if (0) in TRC gives cond. exp. constant warning
#pragma warning(disable: 4127)
#endif

#if	!defined(TRACE) || TRACE==0

#define	TRC		if (0)	{
#define ENDTRC	}
#define TRCXX	0
#else

#if	TRACE!=1 && TRACE !=2
#error "Invalid TRACE"
#endif

#if	TRACE==1

#define TRC			if (1) {
#define ENDTRC		}
#define TRCXX	1
#endif

#if	TRACE==2

#define TRC			if (g_Trace) {
#define ENDTRC		}
#define TRCXX	2
#endif

#endif

extern bool g_Trace;
extern bool g_LogOnly;
void SetTrace(bool On);

#define	ShiftByte(x, n)			(((unsigned char) (x)) << (n))
#define NthByte(x, n)			((unsigned char) ((x) >> ((n)*8) & 0xff))
#define TwoBytes(a, b)			(ShiftByte(a, 0) | ShiftByte(b, 8))
#define FourBytes(a, b, c, d)	(ShiftByte(a, 0) | ShiftByte(b, 8) | ShiftByte(c, 16) | ShiftByte(d, 24))

void PrintAndDie(const char *szFormat, ...);
void SetDieArgcArgv(int argc, char **argv);

void Warning(const char *szFormat, ...);

void Log(const char *szFormat, ...);
void SetLogFile(const char *LogFileName);
void UnsetLogFile();
void Logn(char c, unsigned n);

const std::string &MakeTmpFileName(const std::string &Type, std::string &Name);

char *strsave(const char *str); // succeed-or-die strdup()
char *mystrsave(const char *str); // succeed-or-die strdup()
void safestrcpy(char *s, const char *t, size_t n);

void Dump(const char *Format, ...);

double GetPeakMemUseBytes();
double GetMemUseBytes();
double GetRAMSize();
long GetElapsedSecs();
void ValidateHeap(const char *Msg);
void ChkMem();

void *alloc(size_t bytes);
#define myalloc(t, n)	(t *) alloc((size_t) ((n)*sizeof(t)))
#define myfree(p)		free((void *) (p))

#ifndef _MSC_VER
char *strrev(char *str);
#endif

bool streq(const char *s1, const char *s2);

void VStrPrintf(std::string &Str, const char *szFormat, va_list ArgList);
uint64 atouint64(const char *s);
coord_t StrToCoord(const char *s);
const char *uint64tostr(uint64 i);
const char *uint64tostr(uint64 i, char *s);
const char *uint32tostr(uint32 i);
const char *uint32tostr(uint32 i, char *s);
const char *atoh(const char *str);
const char *utoh(unsigned i);
const char *UIntToStr(unsigned i);
const char *SPScoreToStr(sp_score_t Score);
const char *AIScoreToStr(ai_score_t Score);
const char *ScoreToStr(score_t Score);
const char *CoordToStr(coord_t Pos);
const char *Coord64ToStr(gpos_t Pos);
const char *FileOffsetToStr(off64_t Pos);
unsigned off64_t_to_unsigned(off64_t x);
unsigned uint64_to_unsigned(uint64 x);
coord_t uint64_to_coord_t(uint64 x);
unsigned gpos_t_to_unsigned(gpos_t x);
coord_t gpos_t_to_coord_t(gpos_t Pos);
uint32 coord64_t_to_uint32(gpos_t x);
char *RotatingStaticBuffer();

unsigned DoubleToUInt(double Value);
double UIntToDouble(unsigned Value);

// pom=plus or minus
static inline char pom(bool Plus)
	{
	return Plus ? '+' : '-';
	}

static inline char pomi(int Plus)
	{
	return Plus ? '+' : '-';
	}

// tof=true or false
static inline char tof(bool x)
	{
	return x ? 'T' : 'F';
	}

static inline char tofi(int x)
	{
	return x ? 'T' : 'F';
	}

uint64 pow4(unsigned i);
uint64 pow20(unsigned i);
double Pct(double x, double T);
void OSInit();

bool IsTerminal(int fd);

extern FILE *g_fLog;

#define	SIZE(container)			(unsigned) (container).size()
#define	SIZE32(container)		(uint32) (container).size()
#define FORIX(x)				for (unsigned x##Index = 0; x##Index < x##Count; ++x##Index)
#define FORC(x)					const unsigned x##Count = SIZE(x##s); FORIX(x)

#endif	// UTILS_H
